"""ThreatConnect TI Security Label"""
# first-party
from tcex.utils import Utils

from ..tcex_ti_tc_request import TiTcRequest


class Tag:
    """Unique API calls for Tag API Endpoints

    Args:
        group_type (str): The ThreatConnect define Group type.
        name (str): The name for this Group.
        xid (str, kwargs): The external id for this Group.

    """

    def __init__(self, tcex, name):
        """Initialize Class Properties."""
        self._name = name
        self._tcex = tcex
        self._type = 'tags'
        self._api_sub_type = None
        self._api_type = None
        self._api_entity = 'tag'

        self._utils = Utils()
        self._tc_requests = TiTcRequest(self._tcex)

    @staticmethod
    def is_tag():
        """Return true is instance is a tag object."""
        return True

    def groups(self, group_type=None, filters=None, owner=None, params=None):
        """Get  all groups from a tag.

        Args:
            filters:
            params:
            group_type:
        """
        if group_type and group_type.lower() == 'task':
            group = self._tcex.ti.task()
        else:
            group = self._tcex.ti.group(group_type)
        return self.tc_requests.groups_from_tag(
            group, self.name, filters=filters, owner=owner, params=params
        )

    def indicators(self, indicator_type=None, filters=None, owner=None, params=None):
        """Get all indicators from a tag.

        Args:
            params:
            filters:
            indicator_type:
        """
        indicator = self._tcex.ti.indicator(indicator_type)
        yield from self.tc_requests.indicators_from_tag(
            indicator, self.name, filters=filters, owner=owner, params=params
        )

    def victims(self, filters=None, owner=None, params=None):
        """Get  all victims from a tag."""
        victim = self._tcex.ti.victim()
        yield from self.tc_requests.victims_from_tag(
            victim, self.name, filters=filters, owner=owner, params=params
        )

    @property
    def name(self):
        """Get the tag name."""
        return self._name

    @property
    def tc_requests(self):
        """Get the tc request object"""
        return self._tc_requests

    @name.setter
    def name(self, name):
        """Set  the tag name

        Args:
            name:

        Returns:

        """
        self._name = name

    @tc_requests.setter
    def tc_requests(self, tc_requests):
        """Set the tc request object.

        Args:
            tc_requests:

        Returns:

        """
        self._tc_requests = tc_requests



"""TcEx Mitre Attack Utilities Module"""
# standard library
from typing import List

# flake8: noqa
MITRE_ATTACK_TAGS = """T1001 - Data Obfuscation - C&C - ENT - ATT&CK
T1003 - OS Credential Dumping - CRA - ENT - ATT&CK
T1005 - Data from Local System - COL - ENT - ATT&CK
T1006 - Direct Volume Access - DEF - ENT - ATT&CK
T1007 - System Service Discovery - DIS - ENT - ATT&CK
T1008 - Fallback Channels - C&C - ENT - ATT&CK
T1010 - Application Window Discovery - DIS - ENT - ATT&CK
T1011 - Exfiltration Over Other Network Medium - EXF - ENT - ATT&CK
T1012 - Query Registry - DIS - ENT - ATT&CK
T1014 - Rootkit - DEF - ENT - ATT&CK
T1016 - System Network Configuration Discovery - DIS - ENT - ATT&CK
T1018 - Remote System Discovery - DIS - ENT - ATT&CK
T1020 - Automated Exfiltration - EXF - ENT - ATT&CK
T1021 - Remote Services - LAT - ENT - ATT&CK
T1021.001 - Remote Desktop Protocol - LAT - ENT - ATT&CK
T1021.002 - SMB/Windows Admin Shares - LAT - ENT - ATT&CK
T1021.006 - Windows Remote Management - EXE - ENT - ATT&CK
T1021.006 - Windows Remote Management - LAT - ENT - ATT&CK
T1021.006 - Windows Remote Management - NDT - ENT - ATT&CK
T1025 - Data from Removable Media - COL - ENT - ATT&CK
T1027 - Obfuscated Files or Information - DEF - ENT - ATT&CK
T1027.001 - Binary Padding - DEF - ENT - ATT&CK
T1027.002 - Software Packing - DEF - ENT - ATT&CK
T1027.004 - Compile After Delivery - DEF - ENT - ATT&CK
T1027.005 - Indicator Removal from Tools - DEF - ENT - ATT&CK
T1029 - Scheduled Transfer - EXF - ENT - ATT&CK
T1030 - Data Transfer Size Limits - EXF - ENT - ATT&CK
T1033 - System Owner/User Discovery - DIS - ENT - ATT&CK
T1036 - Masquerading - DEF - ENT - ATT&CK
T1036.006 - Space after Filename - DEF - ENT - ATT&CK
T1036.006 - Space after Filename - EXE - ENT - ATT&CK
T1036.006 - Space after Filename - NDT - ENT - ATT&CK
T1037 - Boot or Logon Initialization Scripts - LAT - ENT - ATT&CK
T1037 - Boot or Logon Initialization Scripts - PER - ENT - ATT&CK
T1037 - Boot or Logon Initialization Scripts - NDT - ENT - ATT&CK
T1037.004 - Rc.common - PER - ENT - ATT&CK
T1037.005 - Startup Items - PER - ENT - ATT&CK
T1037.005 - Startup Items - PRI - ENT - ATT&CK
T1037.005 - Startup Items - NDT - ENT - ATT&CK
T1039 - Data from Network Shared Drive - COL - ENT - ATT&CK
T1040 - Network Sniffing - CRA - ENT - ATT&CK
T1040 - Network Sniffing - DIS - ENT - ATT&CK
T1040 - Network Sniffing - NDT - ENT - ATT&CK
T1041 - Exfiltration Over C2 Channel - EXF - ENT - ATT&CK
T1046 - Network Service Scanning - DIS - ENT - ATT&CK
T1047 - Windows Management Instrumentation - EXE - ENT - ATT&CK
T1048 - Exfiltration Over Alternative Protocol - EXF - ENT - ATT&CK
T1049 - System Network Connections Discovery - DIS - ENT - ATT&CK
T1052 - Exfiltration Over Physical Medium - EXF - ENT - ATT&CK
T1053 - Scheduled Task/Job - EXE - ENT - ATT&CK
T1053 - Scheduled Task/Job - EXE - ENT - ATT&CK
T1053 - Scheduled Task/Job - PER - ENT - ATT&CK
T1053 - Scheduled Task/Job - PER - ENT - ATT&CK
T1053 - Scheduled Task/Job - PRI - ENT - ATT&CK
T1053 - Scheduled Task/Job - NDT - ENT - ATT&CK
T1055 - Process Injection - DEF - ENT - ATT&CK
T1055 - Process Injection - PRI - ENT - ATT&CK
T1055 - Process Injection - NDT - ENT - ATT&CK
T1055.011 - Extra Window Memory Injection - DEF - ENT - ATT&CK
T1055.011 - Extra Window Memory Injection - PRI - ENT - ATT&CK
T1055.011 - Extra Window Memory Injection - NDT - ENT - ATT&CK
T1055.012 - Process Hollowing - DEF - ENT - ATT&CK
T1055.013 - Process Doppelgänging - DEF - ENT - ATT&CK
T1056 - Input Capture - COL - ENT - ATT&CK
T1056 - Input Capture - CRA - ENT - ATT&CK
T1056 - Input Capture - NDT - ENT - ATT&CK
T1056.002 - GUI Input Capture - CRA - ENT - ATT&CK
T1056.004 - Credential API Hooking - CRA - ENT - ATT&CK
T1056.004 - Credential API Hooking - PER - ENT - ATT&CK
T1056.004 - Credential API Hooking - PRI - ENT - ATT&CK
T1056.004 - Credential API Hooking - NDT - ENT - ATT&CK
T1057 - Process Discovery - DIS - ENT - ATT&CK
T1059 - Command and Scripting Interpreter - EXE - ENT - ATT&CK
T1059.001 - PowerShell - EXE - ENT - ATT&CK
T1059.002 - AppleScript - EXE - ENT - ATT&CK
T1059.002 - AppleScript - LAT - ENT - ATT&CK
T1059.002 - AppleScript - NDT - ENT - ATT&CK
T1068 - Exploitation for Privilege Escalation - PRI - ENT - ATT&CK
T1069 - Permission Groups Discovery - DIS - ENT - ATT&CK
T1070 - Indicator Removal on Host - DEF - ENT - ATT&CK
T1070.003 - Clear Command History - DEF - ENT - ATT&CK
T1070.004 - File Deletion - DEF - ENT - ATT&CK
T1070.005 - Network Share Connection Removal - DEF - ENT - ATT&CK
T1070.006 - Timestomp - DEF - ENT - ATT&CK
T1071 - Application Layer Protocol - C&C - ENT - ATT&CK
T1072 - Software Deployment Tools - EXE - ENT - ATT&CK
T1072 - Software Deployment Tools - LAT - ENT - ATT&CK
T1072 - Software Deployment Tools - LAT - ENT - ATT&CK
T1072 - Software Deployment Tools - NDT - ENT - ATT&CK
T1074 - Data Staged - COL - ENT - ATT&CK
T1078 - Valid Accounts - DEF - ENT - ATT&CK
T1078 - Valid Accounts - INI - ENT - ATT&CK
T1078 - Valid Accounts - PER - ENT - ATT&CK
T1078 - Valid Accounts - PRI - ENT - ATT&CK
T1078 - Valid Accounts - NDT - ENT - ATT&CK
T1080 - Taint Shared Content - LAT - ENT - ATT&CK
T1082 - System Information Discovery - DIS - ENT - ATT&CK
T1083 - File and Directory Discovery - DIS - ENT - ATT&CK
T1087 - Account Discovery - DIS - ENT - ATT&CK
T1090 - Proxy - C&C - ENT - ATT&CK
T1090 - Proxy - DEF - ENT - ATT&CK
T1090 - Proxy - NDT - ENT - ATT&CK
T1090.003 - Multi-hop Proxy - C&C - ENT - ATT&CK
T1090.004 - Domain Fronting - C&C - ENT - ATT&CK
T1091 - Replication Through Removable Media - INI - ENT - ATT&CK
T1091 - Replication Through Removable Media - LAT - ENT - ATT&CK
T1091 - Replication Through Removable Media - NDT - ENT - ATT&CK
T1092 - Communication Through Removable Media - C&C - ENT - ATT&CK
T1095 - Non-Application Layer Protocol - C&C - ENT - ATT&CK
T1095 - Non-Application Layer Protocol - C&C - ENT - ATT&CK
T1095 - Non-Application Layer Protocol - NDT - ENT - ATT&CK
T1098 - Account Manipulation - CRA - ENT - ATT&CK
T1098 - Account Manipulation - PER - ENT - ATT&CK
T1098 - Account Manipulation - NDT - ENT - ATT&CK
T1102 - Web Service - C&C - ENT - ATT&CK
T1102 - Web Service - DEF - ENT - ATT&CK
T1102 - Web Service - NDT - ENT - ATT&CK
T1104 - Multi-Stage Channels - C&C - ENT - ATT&CK
T1105 - Ingress Tool Transfer - C&C - ENT - ATT&CK
T1105 - Ingress Tool Transfer - LAT - ENT - ATT&CK
T1105 - Ingress Tool Transfer - NDT - ENT - ATT&CK
T1106 - Native API - EXE - ENT - ATT&CK
T1110 - Brute Force - CRA - ENT - ATT&CK
T1111 - Two-Factor Authentication Interception - CRA - ENT - ATT&CK
T1112 - Modify Registry - DEF - ENT - ATT&CK
T1113 - Screen Capture - COL - ENT - ATT&CK
T1114 - Email Collection - COL - ENT - ATT&CK
T1115 - Clipboard Data - COL - ENT - ATT&CK
T1119 - Automated Collection - COL - ENT - ATT&CK
T1120 - Peripheral Device Discovery - DIS - ENT - ATT&CK
T1123 - Audio Capture - COL - ENT - ATT&CK
T1124 - System Time Discovery - DIS - ENT - ATT&CK
T1125 - Video Capture - COL - ENT - ATT&CK
T1127 - Trusted Developer Utilities Proxy Execution - DEF - ENT - ATT&CK
T1127 - Trusted Developer Utilities Proxy Execution - EXE - ENT - ATT&CK
T1127 - Trusted Developer Utilities Proxy Execution - NDT - ENT - ATT&CK
T1129 - Shared Modules - EXE - ENT - ATT&CK
T1132 - Data Encoding - C&C - ENT - ATT&CK
T1133 - External Remote Services - INI - ENT - ATT&CK
T1133 - External Remote Services - PER - ENT - ATT&CK
T1133 - External Remote Services - NDT - ENT - ATT&CK
T1134 - Access Token Manipulation - DEF - ENT - ATT&CK
T1134 - Access Token Manipulation - PRI - ENT - ATT&CK
T1134 - Access Token Manipulation - NDT - ENT - ATT&CK
T1134.004 - Parent PID Spoofing - DEF - ENT - ATT&CK
T1134.004 - Parent PID Spoofing - PRI - ENT - ATT&CK
T1134.004 - Parent PID Spoofing - NDT - ENT - ATT&CK
T1134.005 - SID-History Injection - PRI - ENT - ATT&CK
T1135 - Network Share Discovery - DIS - ENT - ATT&CK
T1136 - Create Account - PER - ENT - ATT&CK
T1137 - Office Application Startup - PER - ENT - ATT&CK
T1140 - Deobfuscate/Decode Files or Information - DEF - ENT - ATT&CK
T1176 - Browser Extensions - PER - ENT - ATT&CK
T1185 - Man in the Browser - COL - ENT - ATT&CK
T1187 - Forced Authentication - CRA - ENT - ATT&CK
T1189 - Drive-by Compromise - INI - ENT - ATT&CK
T1190 - Exploit Public-Facing Application - INI - ENT - ATT&CK
T1195 - Supply Chain Compromise - INI - ENT - ATT&CK
T1197 - BITS Jobs - DEF - ENT - ATT&CK
T1197 - BITS Jobs - PER - ENT - ATT&CK
T1197 - BITS Jobs - NDT - ENT - ATT&CK
T1199 - Trusted Relationship - INI - ENT - ATT&CK
T1200 - Hardware Additions - INI - ENT - ATT&CK
T1201 - Password Policy Discovery - DIS - ENT - ATT&CK
T1202 - Indirect Command Execution - DEF - ENT - ATT&CK
T1203 - Exploitation for Client Execution - EXE - ENT - ATT&CK
T1204 - User Execution - EXE - ENT - ATT&CK
T1205 - Traffic Signaling - C&C - ENT - ATT&CK
T1205 - Traffic Signaling - DEF - ENT - ATT&CK
T1205 - Traffic Signaling - PER - ENT - ATT&CK
T1205 - Traffic Signaling - NDT - ENT - ATT&CK
T1207 - Rogue Domain Controller - DEF - ENT - ATT&CK
T1210 - Exploitation of Remote Services - LAT - ENT - ATT&CK
T1211 - Exploitation for Defense Evasion - DEF - ENT - ATT&CK
T1212 - Exploitation for Credential Access - CRA - ENT - ATT&CK
T1213 - Data from Information Repositories - COL - ENT - ATT&CK
T1216 - Signed Script Proxy Execution - DEF - ENT - ATT&CK
T1216 - Signed Script Proxy Execution - EXE - ENT - ATT&CK
T1216 - Signed Script Proxy Execution - NDT - ENT - ATT&CK
T1217 - Browser Bookmark Discovery - DIS - ENT - ATT&CK
T1218 - Signed Binary Proxy Execution - DEF - ENT - ATT&CK
T1218 - Signed Binary Proxy Execution - EXE - ENT - ATT&CK
T1218 - Signed Binary Proxy Execution - NDT - ENT - ATT&CK
T1218.001 - Compiled HTML File - DEF - ENT - ATT&CK
T1218.001 - Compiled HTML File - EXE - ENT - ATT&CK
T1218.001 - Compiled HTML File - NDT - ENT - ATT&CK
T1218.002 - Control Panel - DEF - ENT - ATT&CK
T1218.002 - Control Panel - EXE - ENT - ATT&CK
T1218.002 - Control Panel - NDT - ENT - ATT&CK
T1218.003 - CMSTP - DEF - ENT - ATT&CK
T1218.003 - CMSTP - EXE - ENT - ATT&CK
T1218.003 - CMSTP - NDT - ENT - ATT&CK
T1218.004 - InstallUtil - DEF - ENT - ATT&CK
T1218.004 - InstallUtil - EXE - ENT - ATT&CK
T1218.004 - InstallUtil - NDT - ENT - ATT&CK
T1218.005 - Mshta - DEF - ENT - ATT&CK
T1218.005 - Mshta - EXE - ENT - ATT&CK
T1218.005 - Mshta - NDT - ENT - ATT&CK
T1218.007 - Signed Binary Proxy Execution: Msiexec - DEF - ENT - ATT&CK
T1218.009 - Regsvcs/Regasm - DEF - ENT - ATT&CK
T1218.009 - Regsvcs/Regasm - EXE - ENT - ATT&CK
T1218.009 - Regsvcs/Regasm - NDT - ENT - ATT&CK
T1218.010 - Regsvr32 - DEF - ENT - ATT&CK
T1218.010 - Regsvr32 - EXE - ENT - ATT&CK
T1218.010 - Regsvr32 - NDT - ENT - ATT&CK
T1218.011 - Rundll32 - DEF - ENT - ATT&CK
T1218.011 - Rundll32 - EXE - ENT - ATT&CK
T1218.011 - Rundll32 - NDT - ENT - ATT&CK
T1219 - Remote Access Software - C&C - ENT - ATT&CK
T1220 - XSL Script Processing - DEF - ENT - ATT&CK
T1220 - XSL Script Processing - EXE - ENT - ATT&CK
T1220 - XSL Script Processing - NDT - ENT - ATT&CK
T1221 - Template Injection - DEF - ENT - ATT&CK
T1222 - File and Directory Permissions Modification - DEF - ENT - ATT&CK
T1480 - Execution Guardrails - DEF - ENT - ATT&CK
T1482 - Domain Trust Discovery - DIS - ENT - ATT&CK
T1484 - Group Policy Modification - DEF - ENT - ATT&CK
T1485 - Data Destruction - IMP - ENT - ATT&CK
T1486 - Data Encrypted for Impact - IMP - ENT - ATT&CK
T1489 - Service Stop - IMP - ENT - ATT&CK
T1490 - Inhibit System Recovery - IMP - ENT - ATT&CK
T1491 - Defacement - IMP - ENT - ATT&CK
T1495 - Firmware Corruption - IMP - ENT - ATT&CK
T1496 - Resource Hijacking - IMP - ENT - ATT&CK
T1497 - Virtualization/Sandbox Evasion - DEF - ENT - ATT&CK
T1497 - Virtualization/Sandbox Evasion - DIS - ENT - ATT&CK
T1497 - Virtualization/Sandbox Evasion - NDT - ENT - ATT&CK
T1498 - Network Denial of Service - IMP - ENT - ATT&CK
T1499 - Endpoint Denial of Service - IMP - ENT - ATT&CK
T1505 - Server Software Component - PER - ENT - ATT&CK
T1505.003 - Web Shell - PER - ENT - ATT&CK
T1505.003 - Web Shell - PRI - ENT - ATT&CK
T1505.003 - Web Shell - NDT - ENT - ATT&CK
T1518 - Software Discovery - DIS - ENT - ATT&CK
T1518.001 - Security Software Discovery - DIS - ENT - ATT&CK
T1525 - Implant Container Image - PER - ENT - ATT&CK
T1526 - Cloud Service Discovery - DIS - ENT - ATT&CK
T1528 - Steal Application Access Token - CRA - ENT - ATT&CK
T1529 - System Shutdown/Reboot - IMP - ENT - ATT&CK
T1530 - Data from Cloud Storage Object - COL - ENT - ATT&CK
T1531 - Account Access Removal - IMP - ENT - ATT&CK
T1534 - Internal Spearphishing - LAT - ENT - ATT&CK
T1535 - Unused/Unsupported Cloud Regions - DEF - ENT - ATT&CK
T1537 - Transfer Data to Cloud Account - EXF - ENT - ATT&CK
T1538 - Cloud Service Dashboard - DIS - ENT - ATT&CK
T1539 - Steal Web Session Cookie - CRA - ENT - ATT&CK
T1542.001 - System Firmware - PER - ENT - ATT&CK
T1542.002 - Component Firmware - DEF - ENT - ATT&CK
T1542.002 - Component Firmware - PER - ENT - ATT&CK
T1542.002 - Component Firmware - NDT - ENT - ATT&CK
T1542.003 - Bootkit - PER - ENT - ATT&CK
T1543.001 - Launch Agent - PER - ENT - ATT&CK
T1543.002 - Systemd Service - PER - ENT - ATT&CK
T1543.003 - Windows Service - PER - ENT - ATT&CK
T1543.003 - Windows Service - PER - ENT - ATT&CK
T1543.003 - Windows Service - PRI - ENT - ATT&CK
T1543.003 - Windows Service - NDT - ENT - ATT&CK
T1543.004 - Launch Daemon - PER - ENT - ATT&CK
T1543.004 - Launch Daemon - PRI - ENT - ATT&CK
T1543.004 - Launch Daemon - NDT - ENT - ATT&CK
T1546.001 - Change Default File Association - PER - ENT - ATT&CK
T1546.002 - Screensaver - PER - ENT - ATT&CK
T1546.003 - Windows Management Instrumentation Event Subscription - PER - ENT - ATT&CK
T1546.004 - .bash_profile and .bashrc - PER - ENT - ATT&CK
T1546.005 - Trap - EXE - ENT - ATT&CK
T1546.005 - Trap - PER - ENT - ATT&CK
T1546.005 - Trap - NDT - ENT - ATT&CK
T1546.006 - LC_LOAD_DYLIB Addition - PER - ENT - ATT&CK
T1546.007 - Netsh Helper DLL - PER - ENT - ATT&CK
T1546.008 - Accessibility Features - PER - ENT - ATT&CK
T1546.008 - Accessibility Features - PRI - ENT - ATT&CK
T1546.008 - Accessibility Features - NDT - ENT - ATT&CK
T1546.009 - AppCert DLLs - PER - ENT - ATT&CK
T1546.009 - AppCert DLLs - PRI - ENT - ATT&CK
T1546.009 - AppCert DLLs - NDT - ENT - ATT&CK
T1546.010 - AppInit DLLs - PER - ENT - ATT&CK
T1546.010 - AppInit DLLs - PRI - ENT - ATT&CK
T1546.010 - AppInit DLLs - NDT - ENT - ATT&CK
T1546.011 - Application Shimming - PER - ENT - ATT&CK
T1546.011 - Application Shimming - PRI - ENT - ATT&CK
T1546.011 - Application Shimming - NDT - ENT - ATT&CK
T1546.012 - Image File Execution Options Injection - DEF - ENT - ATT&CK
T1546.012 - Image File Execution Options Injection - PER - ENT - ATT&CK
T1546.012 - Image File Execution Options Injection - PRI - ENT - ATT&CK
T1546.012 - Image File Execution Options Injection - NDT - ENT - ATT&CK
T1546.013 - PowerShell Profile - PER - ENT - ATT&CK
T1546.013 - PowerShell Profile - PRI - ENT - ATT&CK
T1546.013 - PowerShell Profile - NDT - ENT - ATT&CK
T1546.014 - Emond - PER - ENT - ATT&CK
T1546.014 - Emond - PRI - ENT - ATT&CK
T1546.014 - Emond - NDT - ENT - ATT&CK
T1546.015 - Component Object Model Hijacking - DEF - ENT - ATT&CK
T1546.015 - Component Object Model Hijacking - PER - ENT - ATT&CK
T1546.015 - Component Object Model Hijacking - NDT - ENT - ATT&CK
T1547.001 - Registry Run Keys / Startup Folder - PER - ENT - ATT&CK
T1547.002 - Authentication Package - PER - ENT - ATT&CK
T1547.003 - Time Providers - PER - ENT - ATT&CK
T1547.004 - Winlogon Helper DLL - PER - ENT - ATT&CK
T1547.005 - Security Support Provider - PER - ENT - ATT&CK
T1547.006 - Kernel Modules and Extensions - PER - ENT - ATT&CK
T1547.007 - Re-opened Applications - PER - ENT - ATT&CK
T1547.008 - LSASS Driver - EXE - ENT - ATT&CK
T1547.008 - LSASS Driver - PER - ENT - ATT&CK
T1547.008 - LSASS Driver - NDT - ENT - ATT&CK
T1547.009 - Shortcut Modification - PER - ENT - ATT&CK
T1547.010 - Port Monitors - PER - ENT - ATT&CK
T1547.010 - Port Monitors - PRI - ENT - ATT&CK
T1547.010 - Port Monitors - NDT - ENT - ATT&CK
T1547.011 - Plist Modification - DEF - ENT - ATT&CK
T1547.011 - Plist Modification - PER - ENT - ATT&CK
T1547.011 - Plist Modification - PER - ENT - ATT&CK
T1547.011 - Plist Modification - PRI - ENT - ATT&CK
T1547.011 - Plist Modification - NDT - ENT - ATT&CK
T1548.001 - Setuid and Setgid - PER - ENT - ATT&CK
T1548.001 - Setuid and Setgid - PRI - ENT - ATT&CK
T1548.001 - Setuid and Setgid - NDT - ENT - ATT&CK
T1548.002 - Bypass User Access Control - DEF - ENT - ATT&CK
T1548.002 - Bypass User Access Control - PRI - ENT - ATT&CK
T1548.002 - Bypass User Access Control - NDT - ENT - ATT&CK
T1548.003 - Sudo and Sudo Caching - PRI - ENT - ATT&CK
T1548.003 - Sudo and Sudo Caching - PRI - ENT - ATT&CK
T1548.003 - Sudo and Sudo Caching - NDT - ENT - ATT&CK
T1548.004 - Elevated Execution with Prompt - PRI - ENT - ATT&CK
T1550.001 - Application Access Token - DEF - ENT - ATT&CK
T1550.001 - Application Access Token - LAT - ENT - ATT&CK
T1550.001 - Application Access Token - NDT - ENT - ATT&CK
T1550.002 - Pass the Hash - LAT - ENT - ATT&CK
T1550.003 - Pass the Ticket - LAT - ENT - ATT&CK
T1550.004 - Web Session Cookie - DEF - ENT - ATT&CK
T1550.004 - Web Session Cookie - LAT - ENT - ATT&CK
T1550.004 - Web Session Cookie - NDT - ENT - ATT&CK
T1552.001 - Credentials In Files - CRA - ENT - ATT&CK
T1552.002 - Credentials in Registry - CRA - ENT - ATT&CK
T1552.003 - Bash History - CRA - ENT - ATT&CK
T1552.004 - Private Keys - CRA - ENT - ATT&CK
T1552.005 - Cloud Instance Metadata API - CRA - ENT - ATT&CK
T1553.001 - Gatekeeper Bypass - DEF - ENT - ATT&CK
T1553.002 - Code Signing - DEF - ENT - ATT&CK
T1553.003 - SIP and Trust Provider Hijacking - DEF - ENT - ATT&CK
T1553.003 - SIP and Trust Provider Hijacking - PER - ENT - ATT&CK
T1553.003 - SIP and Trust Provider Hijacking - NDT - ENT - ATT&CK
T1553.004 - Install Root Certificate - DEF - ENT - ATT&CK
T1555.001 - Keychain - CRA - ENT - ATT&CK
T1555.002 - Securityd Memory - CRA - ENT - ATT&CK
T1555.003 - Credentials from Web Browsers - CRA - ENT - ATT&CK
T1556.002 - Password Filter DLL - CRA - ENT - ATT&CK
T1557.001 - LLMNR/NBT-NS Poisoning and SMB Relay - CRA - ENT - ATT&CK
T1558.003 - Kerberoasting - CRA - ENT - ATT&CK
T1559.002 - Dynamic Data Exchange - EXE - ENT - ATT&CK
T1560 - Archive Collected Data - EXF - ENT - ATT&CK
T1560 - Archive Collected Data - EXF - ENT - ATT&CK
T1560 - Archive Collected Data - NDT - ENT - ATT&CK
T1561.001 - Disk Content Wipe - IMP - ENT - ATT&CK
T1561.002 - Disk Structure Wipe - IMP - ENT - ATT&CK
T1562.001 - Disable or Modify Tools - DEF - ENT - ATT&CK
T1562.003 - HISTCONTROL - DEF - ENT - ATT&CK
T1562.006 - Indicator Blocking - DEF - ENT - ATT&CK
T1563.001 - SSH Hijacking - LAT - ENT - ATT&CK
T1564.001 - Hidden Files and Directories - DEF - ENT - ATT&CK
T1564.001 - Hidden Files and Directories - PER - ENT - ATT&CK
T1564.001 - Hidden Files and Directories - NDT - ENT - ATT&CK
T1564.002 - Hidden Users - DEF - ENT - ATT&CK
T1564.003 - Hidden Window - DEF - ENT - ATT&CK
T1564.004 - NTFS File Attributes - DEF - ENT - ATT&CK
T1565.001 - Stored Data Manipulation - IMP - ENT - ATT&CK
T1565.002 - Transmitted Data Manipulation - IMP - ENT - ATT&CK
T1565.003 - Runtime Data Manipulation - IMP - ENT - ATT&CK
T1566.001 - Spearphishing Attachment - INI - ENT - ATT&CK
T1566.002 - Spearphishing Link - INI - ENT - ATT&CK
T1566.003 - Spearphishing via Service - INI - ENT - ATT&CK
T1568.002 - Domain Generation Algorithms - C&C - ENT - ATT&CK
T1569.001 - Launchctl - DEF - ENT - ATT&CK
T1569.001 - Launchctl - EXE - ENT - ATT&CK
T1569.001 - Launchctl - PER - ENT - ATT&CK
T1569.001 - Launchctl - NDT - ENT - ATT&CK
T1569.002 - Service Execution - EXE - ENT - ATT&CK
T1571 - Non-Standard Port - C&C - ENT - ATT&CK
T1573 - Encrypted Channel - C&C - ENT - ATT&CK
T1573 - Encrypted Channel - C&C - ENT - ATT&CK
T1573 - Encrypted Channel - C&C - ENT - ATT&CK
T1573 - Encrypted Channel - NDT - ENT - ATT&CK
T1574.001 - DLL Search Order Hijacking - DEF - ENT - ATT&CK
T1574.001 - DLL Search Order Hijacking - PER - ENT - ATT&CK
T1574.001 - DLL Search Order Hijacking - PRI - ENT - ATT&CK
T1574.001 - DLL Search Order Hijacking - NDT - ENT - ATT&CK
T1574.002 - DLL Side-Loading - DEF - ENT - ATT&CK
T1574.004 - Dylib Hijacking - PER - ENT - ATT&CK
T1574.004 - Dylib Hijacking - PRI - ENT - ATT&CK
T1574.004 - Dylib Hijacking - NDT - ENT - ATT&CK
T1574.010 - Services File Permissions Weakness - PER - ENT - ATT&CK
T1574.010 - Services File Permissions Weakness - PRI - ENT - ATT&CK
T1574.010 - Services File Permissions Weakness - NDT - ENT - ATT&CK
T1574.011 - Services Registry Permissions Weakness - PER - ENT - ATT&CK
T1574.011 - Services Registry Permissions Weakness - PRI - ENT - ATT&CK
T1574.011 - Services Registry Permissions Weakness - NDT - ENT - ATT&CK
T1578.004 - Revert Cloud Instance - DEF - ENT - ATT&CK
T1573.001 - Symmetric Cryptography - C&C - ENT - ATT&CK
T1052.001 - Exfiltration over USB - EXF - ENT - ATT&CK
T1560.001 - Archive via Utility - EXF - ENT - ATT&CK
T1071.001 - Web Protocols - C&C - ENT - ATT&CK
T1224 - Assess leadership areas of interest - PDP - PRE - ATT&CK
T1225 - Identify gap areas - PDP - PRE - ATT&CK
T1226 - Conduct cost/benefit analysis - PDP - PRE - ATT&CK
T1227 - Develop KITs/KIQs - PDP - PRE - ATT&CK
T1228 - Assign KITs/KIQs into categories - PDP - PRE - ATT&CK
T1229 - Assess KITs/KIQs benefits - PDP - PRE - ATT&CK
T1230 - Derive intelligence requirements - PDP - PRE - ATT&CK
T1231 - Create strategic plan - PDP - PRE - ATT&CK
T1232 - Create implementation plan - PDP - PRE - ATT&CK
T1233 - Identify analyst level gaps - PDP - PRE - ATT&CK
T1234 - Generate analyst intelligence requirements - PDP - PRE - ATT&CK
T1235 - Receive operator KITs/KIQs tasking - PDP - PRE - ATT&CK
T1236 - Assess current holdings, needs, and wants - PDP - PRE - ATT&CK
T1237 - Submit KITs, KIQs, and intelligence requirements - PDD - PRE - ATT&CK
T1238 - Assign KITs, KIQs, and/or intelligence requirements - PDD - PRE - ATT&CK
T1239 - Receive KITs/KIQs and determine requirements - PDD - PRE - ATT&CK
T1240 - Task requirements - PDD - PRE - ATT&CK
T1241 - Determine strategic target - TAR - PRE - ATT&CK
T1242 - Determine operational element - TAR - PRE - ATT&CK
T1243 - Determine highest level tactical element - TAR - PRE - ATT&CK
T1244 - Determine secondary level tactical element - TAR - PRE - ATT&CK
T1245 - Determine approach/attack vector - TAR - PRE - ATT&CK
T1246 - Identify supply chains - TIG - PRE - ATT&CK
T1247 - Acquire OSINT data sets and information - TIG - PRE - ATT&CK
T1248 - Identify job postings and needs/gaps - TIG - PRE - ATT&CK
T1249 - Conduct social engineering - TIG - PRE - ATT&CK
T1250 - Determine domain and IP address space - TIG - PRE - ATT&CK
T1251 - Obtain domain/IP registration information - TIG - PRE - ATT&CK
T1252 - Map network topology - TIG - PRE - ATT&CK
T1253 - Conduct passive scanning - TIG - PRE - ATT&CK
T1254 - Conduct active scanning - TIG - PRE - ATT&CK
T1255 - Discover target logon/email address format - TIG - PRE - ATT&CK
T1256 - Identify web defensive services - TIG - PRE - ATT&CK
T1257 - Mine technical blogs/forums - TIG - PRE - ATT&CK
T1258 - Determine firmware version - TIG - PRE - ATT&CK
T1259 - Determine external network trust dependencies - TIG - PRE - ATT&CK
T1260 - Determine 3rd party infrastructure services - TIG - PRE - ATT&CK
T1261 - Enumerate externally facing software applications technologies, languages, and dependencies - TIG - PRE - ATT&CK
T1262 - Enumerate client configurations - TIG - PRE - ATT&CK
T1263 - Identify security defensive capabilities - TIG - PRE - ATT&CK
T1264 - Identify technology usage patterns - TIG - PRE - ATT&CK
T1265 - Identify supply chains - PIG - PRE - ATT&CK
T1266 - Acquire OSINT data sets and information - PIG - PRE - ATT&CK
T1267 - Identify job postings and needs/gaps - PIG - PRE - ATT&CK
T1268 - Conduct social engineering - PIG - PRE - ATT&CK
T1269 - Identify people of interest - PIG - PRE - ATT&CK
T1270 - Identify groups/roles - PIG - PRE - ATT&CK
T1271 - Identify personnel with an authority/privilege - PIG - PRE - ATT&CK
T1272 - Identify business relationships - PIG - PRE - ATT&CK
T1273 - Mine social media - PIG - PRE - ATT&CK
T1274 - Identify sensitive personnel information - PIG - PRE - ATT&CK
T1275 - Aggregate individual's digital footprint - PIG - PRE - ATT&CK
T1276 - Identify supply chains - OIG - PRE - ATT&CK
T1277 - Acquire OSINT data sets and information - OIG - PRE - ATT&CK
T1278 - Identify job postings and needs/gaps - OIG - PRE - ATT&CK
T1279 - Conduct social engineering - OIG - PRE - ATT&CK
T1280 - Identify business processes/tempo - OIG - PRE - ATT&CK
T1281 - Obtain templates/branding materials - OIG - PRE - ATT&CK
T1282 - Determine physical locations - OIG - PRE - ATT&CK
T1283 - Identify business relationships - OIG - PRE - ATT&CK
T1284 - Determine 3rd party infrastructure services - OIG - PRE - ATT&CK
T1285 - Determine centralization of IT management - OIG - PRE - ATT&CK
T1286 - Dumpster dive - OIG - PRE - ATT&CK
T1287 - Analyze data collected - TWI - PRE - ATT&CK
T1288 - Analyze architecture and configuration posture - TWI - PRE - ATT&CK
T1289 - Analyze organizational skillsets and deficiencies - TWI - PRE - ATT&CK
T1290 - Research visibility gap of security vendors - TWI - PRE - ATT&CK
T1291 - Research relevant vulnerabilities/CVEs - TWI - PRE - ATT&CK
T1292 - Test signature detection - TWI - PRE - ATT&CK
T1293 - Analyze application security posture - TWI - PRE - ATT&CK
T1294 - Analyze hardware/software security defensive capabilities - TWI - PRE - ATT&CK
T1295 - Analyze social and business relationships, interests, and affiliations - PWI - PRE - ATT&CK
T1296 - Assess targeting options - PWI - PRE - ATT&CK
T1297 - Analyze organizational skillsets and deficiencies - PWI - PRE - ATT&CK
T1298 - Assess vulnerability of 3rd party vendors - OWI - PRE - ATT&CK
T1299 - Assess opportunities created by business deals - OWI - PRE - ATT&CK
T1300 - Analyze organizational skillsets and deficiencies - OWI - PRE - ATT&CK
T1301 - Analyze business processes - OWI - PRE - ATT&CK
T1302 - Assess security posture of physical locations - OWI - PRE - ATT&CK
T1303 - Analyze presence of outsourced capabilities - OWI - PRE - ATT&CK
T1304 - Proxy/protocol relays - AOP - PRE - ATT&CK
T1305 - Private whois services - AOP - PRE - ATT&CK
T1306 - Anonymity services - AOP - PRE - ATT&CK
T1307 - Acquire and/or use 3rd party infrastructure services - AOP - PRE - ATT&CK
T1308 - Acquire and/or use 3rd party software services - AOP - PRE - ATT&CK
T1309 - Obfuscate infrastructure - AOP - PRE - ATT&CK
T1310 - Acquire or compromise 3rd party signing certificates - AOP - PRE - ATT&CK
T1311 - Dynamic DNS - AOP - PRE - ATT&CK
T1312 - Compromise 3rd party infrastructure to support delivery - AOP - PRE - ATT&CK
T1313 - Obfuscation or cryptography - AOP - PRE - ATT&CK
T1314 - Host-based hiding techniques - AOP - PRE - ATT&CK
T1315 - Network-based hiding techniques - AOP - PRE - ATT&CK
T1316 - Non-traditional or less attributable payment options - AOP - PRE - ATT&CK
T1317 - Secure and protect infrastructure - AOP - PRE - ATT&CK
T1318 - Obfuscate operational infrastructure - AOP - PRE - ATT&CK
T1319 - Obfuscate or encrypt code - AOP - PRE - ATT&CK
T1320 - Data Hiding - AOP - PRE - ATT&CK
T1321 - Common, high volume protocols and software - AOP - PRE - ATT&CK
T1322 - Misattributable credentials - AOP - PRE - ATT&CK
T1323 - Domain Generation Algorithms (DGA) - AOP - PRE - ATT&CK
T1324 - DNSCalc - AOP - PRE - ATT&CK
T1325 - Fast Flux DNS - AOP - PRE - ATT&CK
T1326 - Domain registration hijacking - EMI - PRE - ATT&CK
T1327 - Use multiple DNS infrastructures - EMI - PRE - ATT&CK
T1328 - Buy domain name - EMI - PRE - ATT&CK
T1329 - Acquire and/or use 3rd party infrastructure services - EMI - PRE - ATT&CK
T1330 - Acquire and/or use 3rd party software services - EMI - PRE - ATT&CK
T1331 - Obfuscate infrastructure - EMI - PRE - ATT&CK
T1332 - Acquire or compromise 3rd party signing certificates - EMI - PRE - ATT&CK
T1333 - Dynamic DNS - EMI - PRE - ATT&CK
T1334 - Compromise 3rd party infrastructure to support delivery - EMI - PRE - ATT&CK
T1335 - Procure required equipment and software - EMI - PRE - ATT&CK
T1336 - Install and configure hardware, network, and systems - EMI - PRE - ATT&CK
T1337 - SSL certificate acquisition for domain - EMI - PRE - ATT&CK
T1338 - SSL certificate acquisition for trust breaking - EMI - PRE - ATT&CK
T1339 - Create backup infrastructure - EMI - PRE - ATT&CK
T1340 - Shadow DNS - EMI - PRE - ATT&CK
T1341 - Build social network persona - PDV - PRE - ATT&CK
T1342 - Develop social network persona digital footprint - PDV - PRE - ATT&CK
T1343 - Choose pre-compromised persona and affiliated accounts - PDV - PRE - ATT&CK
T1344 - Friend/Follow/Connect to targets of interest - PDV - PRE - ATT&CK
T1345 - Create custom payloads - BDC - PRE - ATT&CK
T1346 - Obtain/re-use payloads - BDC - PRE - ATT&CK
T1347 - Build and configure delivery systems - BDC - PRE - ATT&CK
T1348 - Identify resources required to build capabilities - BDC - PRE - ATT&CK
T1349 - Build or acquire exploits - BDC - PRE - ATT&CK
T1350 - Discover new exploits and monitor exploit-provider forums - BDC - PRE - ATT&CK
T1351 - Remote access tool development - BDC - PRE - ATT&CK
T1352 - C2 protocol development - BDC - PRE - ATT&CK
T1353 - Post compromise tool development - BDC - PRE - ATT&CK
T1354 - Compromise 3rd party or closed-source vulnerability/exploit information - BDC - PRE - ATT&CK
T1355 - Create infected removable media - BDC - PRE - ATT&CK
T1356 - Test callback functionality - TST - PRE - ATT&CK
T1357 - Test malware in various execution environments - TST - PRE - ATT&CK
T1358 - Review logs and residual traces - TST - PRE - ATT&CK
T1359 - Test malware to evade detection - TST - PRE - ATT&CK
T1360 - Test physical access - TST - PRE - ATT&CK
T1361 - Test signature detection for file upload/email filters - TST - PRE - ATT&CK
T1362 - Upload, install, and configure software/tools - STG - PRE - ATT&CK
T1363 - Port redirector - STG - PRE - ATT&CK
T1364 - Friend/Follow/Connect to targets of interest - STG - PRE - ATT&CK
T1365 - Hardware or software supply chain implant - STG - PRE - ATT&CK
T1366 - Targeted social media phishing - LNC - PRE - ATT&CK
T1367 - Spear phishing messages with malicious attachments - LNC - PRE - ATT&CK
T1368 - Spear phishing messages with text only - LNC - PRE - ATT&CK
T1369 - Spear phishing messages with malicious links - LNC - PRE - ATT&CK
T1370 - Untargeted client-side exploitation - LNC - PRE - ATT&CK
T1371 - Targeted client-side exploitation - LNC - PRE - ATT&CK
T1372 - Unconditional client-side exploitation/Injected Website/Driveby - LNC - PRE - ATT&CK
T1373 - Push-notification client-side exploit - LNC - PRE - ATT&CK
T1374 - Credential pharming - LNC - PRE - ATT&CK
T1375 - Leverage compromised 3rd party resources - LNC - PRE - ATT&CK
T1376 - Conduct social engineering or HUMINT operation - LNC - PRE - ATT&CK
T1377 - Exploit public-facing application - LNC - PRE - ATT&CK
T1378 - Replace legitimate binary with malware - LNC - PRE - ATT&CK
T1379 - Disseminate removable media - STG - PRE - ATT&CK
T1380 - Deploy exploit using advertising - LNC - PRE - ATT&CK
T1381 - Authentication attempt - LNC - PRE - ATT&CK
T1382 - DNS poisoning - LNC - PRE - ATT&CK
T1383 - Confirmation of launched compromise achieved - COM - PRE - ATT&CK
T1384 - Automated system performs requested action - COM - PRE - ATT&CK
T1385 - Human performs requested action of physical nature - COM - PRE - ATT&CK
T1386 - Authorized user performs requested cyber action - COM - PRE - ATT&CK
T1387 - Unauthorized user introduces compromise delivery mechanism - COM - PRE - ATT&CK
T1388 - Compromise of externally facing system - COM - PRE - ATT&CK
T1389 - Identify vulnerabilities in third-party software libraries - TWI - PRE - ATT&CK
T1390 - OS-vendor provided communication channels - AOP - PRE - ATT&CK
T1391 - Choose pre-compromised mobile app developer account credentials or signing keys - PDV - PRE - ATT&CK
T1392 - Obtain Apple iOS enterprise distribution key pair and certificate - PDV - PRE - ATT&CK
T1393 - Test ability to evade automated mobile application security analysis performed by app stores - TST - PRE - ATT&CK
T1394 - Distribute malicious software development tools - STG - PRE - ATT&CK
T1395 - Runtime code download and execution - LNC - PRE - ATT&CK
T1396 - Obtain booter/stressor subscription - EMI - PRE - ATT&CK
T1397 - Spearphishing for Information - TIG - PRE - ATT&CK"""
MITRE_ATTACK_TAGS_SPLIT = MITRE_ATTACK_TAGS.split('\n')
MITRE_ATTACK_TACTIC_ABBREVIATIONS = {
    'adversaryopsec': 'AOP',
    'buildcapabilities': 'BDC',
    'collection': 'COL',
    'commandandcontrol': 'C&C',
    'compromise': 'COM',
    'credentialaccess': 'CRA',
    'defenseevasion': 'DEF',
    'discovery': 'DIS',
    'establish&maintaininfrastructure': 'EMI',
    'execution': 'EXE',
    'exfiltration': 'EXF',
    'impact': 'IMP',
    'initialaccess': 'INI',
    'lateralmovement': 'LAT',
    'launch': 'LNC',
    'organizationalinformationgathering': 'OIG',
    'organizationalweaknessidentification': 'OWI',
    'peopleinformationgathering': 'PIG',
    'peopleweaknessidentification': 'PWI',
    'persistence': 'PER',
    'personadevelopment': 'PDV',
    'prioritydefinitiondirection': 'PDD',
    'prioritydefinitionplanning': 'PDP',
    'privilegeescalation': 'PRI',
    'stagecapabilities': 'STG',
    'targetselection': 'TAR',
    'technicalinformationgathering': 'TIG',
    'technicalweaknessidentification': 'TWI',
    'testcapabilities': 'TST',
    'unknown': 'NDT',
}


class MitreAttackUtils:
    """TcEx framework Mitre Attacks Utils class"""

    @staticmethod
    def technique_id_to_tags(technique_id: str, include_ndt_tag: bool = False) -> List[str]:
        """Return the ThreatConnect tags for the given technique_id."""
        matching_tags = []
        technique_id_prefix = f'{technique_id.title()} -'
        for tag in MITRE_ATTACK_TAGS_SPLIT:
            if tag.startswith(technique_id_prefix):
                is_ndt_tag = '- NDT - ' in tag
                if include_ndt_tag or not is_ndt_tag:
                    matching_tags.append(tag)

        return matching_tags

    @staticmethod
    def tactic_name_abbreviation(tactic_name: str):
        """Return the ThreatConnect abbreviation for the given tactic_name."""
        tactic_name = tactic_name.replace(' ', '').replace('-', '').lower()
        return MITRE_ATTACK_TACTIC_ABBREVIATIONS.get(tactic_name)



#!/usr/bin/env python
"""TcEx Framework LayoutJson."""
# standard library
import json
import logging
import os
import random
import sys

try:
    # standard library
    import sqlite3
except ImportError:
    # only required for local development
    pass

from .install_json import InstallJson
from .layout_json import LayoutJson


class Permutations:
    """Permutations Module

    Args:
        logger (logging.Logger, optional): A instance of Logger. Defaults to None.
    """

    def __init__(self, logger=None):
        """Initialize Class properties"""
        self.log = logger or logging.getLogger('permutations')

        # properties
        self._db_conn = None
        self._input_names = None
        self._input_permutations = None
        self._output_permutations = None
        self.app_path = os.getcwd()
        self.ij = InstallJson()
        self.lj = LayoutJson()
        self.input_table = 'inputs'

    def _gen_permutations(self, index=0, args=None):
        """Iterate recursively over layout.json parameter names to build permutations.

        .. NOTE:: Permutations are for layout.json based Apps.

        Args:
            index (int, optional): The current index position in the layout names list.
            args (list, optional): Defaults to None. The current list of args.
        """
        if args is None:
            args = []
        try:
            hidden = False
            if self.ij.runtime_level.lower() in [
                'playbook',
                'triggerservice',
                'webhooktriggerservice',
            ]:
                name = list(self.lj.parameters_names)[index]
                display = self.lj.params_dict.get(name, {}).get('display')
                hidden = self.lj.params_dict.get(name, {}).get('hidden', False)
            else:
                name = list(self.ij.params_dict.keys())[index]
                display = False

            input_type = self.ij.params_dict.get(name, {}).get('type')
            if input_type is None:
                self.handle_error(f'No value found in install.json for "{name}".')

            if (
                self.ij.runtime_level.lower() == 'organization'
                or self.validate_layout_display(self.input_table, display)
                or hidden
            ):
                if input_type.lower() == 'boolean':
                    for val in [True, False]:
                        args.append({'name': name, 'value': val})
                        self.db_update_record(self.input_table, name, val)
                        self._gen_permutations(index + 1, list(args))
                        # remove the previous arg before next iteration
                        args.pop()
                elif input_type.lower() == 'choice':
                    valid_values = self.ij.expand_valid_values(
                        self.ij.params_dict.get(name, {}).get('validValues', [])
                    )
                    for val in valid_values:
                        args.append({'name': name, 'value': val})
                        self.db_update_record(self.input_table, name, val)
                        self._gen_permutations(index + 1, list(args))
                        # remove the previous arg before next iteration
                        args.pop()
                else:
                    args.append({'name': name, 'value': None})
                    self._gen_permutations(index + 1, list(args))
            else:
                self._gen_permutations(index + 1, list(args))

        except IndexError:
            # when IndexError is reached all data has been processed.
            self._input_permutations.append(args)
            outputs = []

            for output_data in self.ij.output_variables:
                name = output_data.get('name')
                if self.lj.outputs_dict.get(name) is not None:
                    display = self.lj.outputs_dict.get(name, {}).get('display')
                    valid = self.validate_layout_display(self.input_table, display)
                    if display is None or not valid:
                        continue
                outputs.append(output_data)
            self._output_permutations.append(outputs)

    @property
    def db_conn(self):
        """Create a temporary in memory DB and return the connection."""
        if self._db_conn is None:
            try:
                self._db_conn = sqlite3.connect(':memory:')
            except sqlite3.Error as e:
                self.handle_error(e)
        return self._db_conn

    def db_create_table(self, table_name, columns):
        """Create a temporary DB table.

        Arguments:
            table_name (str): The name of the table.
            columns (list): List of columns to add to the DB.
        """
        formatted_columns = ''
        for col in set(columns):
            formatted_columns += f""""{col.strip('"').strip("'")}" text, """
        formatted_columns = formatted_columns.strip(', ')

        create_table_sql = f'CREATE TABLE IF NOT EXISTS {table_name} ({formatted_columns});'
        try:
            cr = self.db_conn.cursor()
            cr.execute(create_table_sql)
        except sqlite3.Error as e:
            self.handle_error(e)

    def db_drop_table(self, table_name):
        """Drop a DB table.

        Arguments:
            table_name (str): The name of the table.
        """
        create_table_sql = f'DROP TABLE IF EXISTS {table_name};'
        try:
            cr = self.db_conn.cursor()
            cr.execute(create_table_sql)
        except sqlite3.Error as e:
            self.handle_error(e)

    def db_insert_record(self, table_name, columns):
        """Insert records into DB.

        Args:
            table_name (str): The name of the table.
            columns (list): List of columns for insert statement.
        """
        bindings = ('?,' * len(columns)).strip(',')
        values = [None] * len(columns)
        try:
            sql = f"INSERT INTO {table_name} ({', '.join(columns)}) VALUES ({bindings})"
            cur = self.db_conn.cursor()
            cur.execute(sql, values)
        except sqlite3.OperationalError as e:
            raise RuntimeError(f'SQL insert failed - SQL: "{sql}", Error: "{e}"')

    def db_update_record(self, table_name, column, value):
        """Insert records into DB.

        Args:
            table_name (str): The name of the table.
            column (str): The column name in which the value is to be updated.
            value (str): The value to update in the column.
        """
        # escape any single quotes in value
        if isinstance(value, str):
            value = value.replace('\'', '\\')
        elif isinstance(value, bool):
            # core expects true/false so we convert bool value to string and lower
            value = str(value).lower()

        # only column defined in install.json can be updated
        if column in self.ij.params_dict:
            try:
                # value should be wrapped in single quotes to be properly parsed
                sql = f"UPDATE {table_name} SET {column} = '{value}'"
                cur = self.db_conn.cursor()
                cur.execute(sql)
            except sqlite3.OperationalError as e:
                raise RuntimeError(f'SQL update failed - SQL: "{sql}", Error: "{e}"')

    def exists(self):
        """Return True if permutation file exists."""
        return os.path.isfile(self.filename)

    @property
    def filename(self):
        """Return all output permutations for current App."""
        return os.path.join(self.app_path, 'permutations.json')

    @staticmethod
    def handle_error(err, halt=True):
        """Print errors message and optionally exit.

        Args:
            err (str): The error message to print.
            halt (bool, optional): Defaults to True. If True the script will exit.
        """
        print(err)
        if halt:
            sys.exit(1)

    def init_permutations(self):
        """Process layout.json names/display to get all permutations of args."""
        if self._input_permutations is None and self._output_permutations is None:
            self._input_permutations = []
            self._output_permutations = []

            # create db for permutations testing
            self.db_create_table(self.input_table, self.ij.params_dict.keys())
            self.db_insert_record(self.input_table, self.ij.params_dict.keys())

            # only gen permutations if none have been generated previously
            self._gen_permutations()

            # drop database
            self.db_drop_table(self.input_table)

    def input_dict(self, permutation_id):
        """Return all input permutation names for provided permutation id.

        {'tc_action': 'Append', 'input_strings': None, 'append_chars': None}

        Args:
            permutation_id (int): The index of the permutation input array.

        Returns:
            dict: A dict with key / value for each input for the provided permutation id.
        """
        input_dict = {}
        if self.lj.has_layout:
            for permutation in self.input_permutations[permutation_id]:
                input_dict.setdefault(permutation.get('name'), permutation.get('value'))
        return input_dict

    @property
    def input_names(self):
        """Return all input permutation names for current App."""
        if self._input_names is None and self.lj.has_layout:
            self._input_names = []
            for permutation in self.input_permutations:
                self._input_names.append([p.get('name') for p in permutation])
        return self._input_names

    @property
    def input_permutations(self):
        """Return all input permutations for current App.

        self._input_permutations is an array of permutations arrays.
        [[<perm obj #1], [<perm obj #2]]
        """
        if self._input_permutations is None and self.lj.has_layout:
            self.init_permutations()
        return self._input_permutations

    @property
    def output_permutations(self):
        """Return all output permutations for current App."""
        if self._output_permutations is None:
            self.init_permutations()
        return self._output_permutations

    def outputs_by_inputs(self, inputs):
        """Return all output based on provided inputs

        Args:
            inputs (dict): The args/inputs dict.
        """
        table = f'temp_{random.randint(100,999)}'  # nosec
        self.db_create_table(table, self.ij.params_dict.keys())
        self.db_insert_record(table, self.ij.params_dict.keys())

        for name, val in inputs.items():
            self.db_update_record(table, name, val)

        outputs = []
        # loop through all output variables in install.json
        for output_data in self.ij.output_variables:
            name = output_data.get('name')
            if self.lj.outputs_dict.get(name) is None:
                # an output not listed in layout.json should always be shown
                valid = True
            else:
                # all other outputs must be validated
                display = self.lj.outputs_dict.get(name, {}).get('display')
                valid = self.validate_layout_display(table, display)

            if valid:
                # valid outputs get added to array
                outputs.append(output_data)

        # drop database
        self.db_drop_table(table)

        return outputs

    def permutations(self):
        """Process layout.json names/display to get all permutations of args."""
        if 'sqlite3' not in sys.modules:
            print('The sqlite3 module needs to be build-in to Python for this feature.')
            sys.exit(1)

        # create db for permutations testing
        self.db_create_table(self.input_table, self.ij.params_dict.keys())
        self.db_insert_record(self.input_table, self.ij.params_dict.keys())

        # only gen permutations if none have been generated previously
        if not self._input_permutations and not self._output_permutations:
            self._gen_permutations()

        # output permutations
        self.write_permutations_file()

    def validate_input_variable(self, input_name, inputs):
        """Return True if the provided variables display where clause returns results.

        Args:
            input_name (dict): The input variable name (e.g. tc_action).
            inputs (dict): The current name/value dict.

        Returns:
            bool: True if the display value returns results.
        """
        if not self.lj.has_layout or not inputs:
            # always return try if current App doesn't have a layouts file
            return True

        table = f'temp_{random.randint(100,999)}'  # nosec
        self.db_create_table(table, self.ij.params_dict.keys())
        self.db_insert_record(table, self.ij.params_dict.keys())

        # APP-98 Added to cover the use case of interdependent variables in the layout.json.
        for name, item in self.ij.filter_params_dict(_type='Boolean').items():
            self.db_update_record(table, name, item.get('default', False))

        for name, val in inputs.items():
            self.db_update_record(table, name, val)

        lj_data = self.lj.params_dict.get(input_name)
        if lj_data is None:
            # this shouldn't happen as all ij inputs must be in lj
            raise RuntimeError(f'The provided input {input_name} was not found in layout.json.')
        display = lj_data.get('display')

        # check if provided variable meets display requirements
        valid = self.validate_layout_display(table, display)

        # cleanup temp table
        self.db_drop_table(table)

        return valid

    def validate_layout_display(self, table, display_condition):
        """Check to see if the display condition passes.

        Args:
            table (str): The name of the DB table which hold the App data.
            display_condition (str): The "where" clause of the DB SQL statement.

        Returns:
            bool: True if the row count is greater than 0.
        """
        display = False
        if display_condition is None:
            display = True
        else:
            display_query = f'select count(*) from {table} where {display_condition}'  # nosec
            try:
                cur = self.db_conn.cursor()
                cur.execute(display_query.replace('"', ''))
                rows = cur.fetchall()
                if rows[0][0] > 0:
                    display = True
            except sqlite3.Error as e:
                print(f'"{display_query}" query returned an error: ({e}).')
                sys.exit(1)
        return display

    def write_permutations_file(self):
        """Print all valid permutations."""
        permutations = []
        for index, p in enumerate(self.input_permutations):
            permutations.append({'index': index, 'args': p})

        with open(self.filename, 'w') as fh:
            json.dump(permutations, fh, indent=2, sort_keys=True)
        print('All permutations written to the "permutations.json" file.')



"""API Handler Class"""
# standard library
import logging
import threading
import time


class ApiHandler(logging.Handler):
    """Logger handler for ThreatConnect Exchange API logging."""

    def __init__(self, session, flush_limit=100):
        """Initialize Class properties.

        Args:
            session (Request.Session): The preconfigured instance of Session for ThreatConnect API.
            flush_limit (int): The limit to flush batch logs to the API.
        """
        super().__init__()
        self.session = session
        self.flush_limit = flush_limit
        self._entries = []
        self.in_token_renewal = False

    def flush(self):
        """Close the logger and flush entries."""
        self.log_to_api(self.entries)  # pragma: no cover

    def emit(self, record):
        """Emit a record.

        Args:
            record (obj): The record to be logged.
        """
        if threading.current_thread().name != 'MainThread':  # pragma: no cover
            return

        # queue log events
        self._entries.append(self.format(record))

        # flush queue once limit is hit token module is currently renewing token
        if (
            len(self._entries) > self.flush_limit or record.levelname == 'ERROR'
        ) and not self.in_token_renewal:
            self.log_to_api(self.entries)

    @property
    def entries(self):
        """Return a copy and clear self._entries."""
        entries = list(self._entries)
        self._entries = []
        return entries

    def log_to_api(self, entries):
        """Send log events to the ThreatConnect API"""
        if entries:
            try:
                headers = {'Content-Type': 'application/json'}
                self.session.post('/v2/logs/app', headers=headers, json=entries)
            except Exception:  # nosec; pragma: no cover
                pass


class ApiHandlerFormatter(logging.Formatter):
    """Logger formatter for ThreatConnect Exchange API logging."""

    def __init__(self):
        """Initialize Class properties."""
        super().__init__()

    def format(self, record):
        """Format log record for ThreatConnect API.

        Example log event::

            [{
                "timestamp": 1478907537000,
                "message": "Test Message",
                "level": "DEBUG"
            }]

        Args:
            record (obj): The record to be logged.
        """
        return {
            'timestamp': int(float(record.created or time.time()) * 1000),
            'message': record.msg or '',
            'level': record.levelname or 'DEBUG',
        }



"""API Handler Class"""
# standard library
import gzip
import os
import shutil
from logging.handlers import RotatingFileHandler
from typing import Optional


class RotatingFileHandlerCustom(RotatingFileHandler):
    """Logger handler for ThreatConnect Exchange File logging."""

    def __init__(
        self,
        filename: str,
        mode: Optional[str] = 'a',
        maxBytes: Optional[int] = 0,
        backupCount: Optional[int] = 0,
        encoding: Optional[str] = None,
        delay: Optional[bool] = False,
    ):
        """Customize RotatingFileHandler to create full log path.

        Args:
            filename: The name of the logfile.
            mode: The write mode for the file.
            maxBytes: The max file size before rotating.
            backupCount: The maximum # of backup files.
            encoding: The log file encoding.
            delay: If True, then file opening is deferred until the first call to emit().
        """
        if encoding is None and os.getenv('LANG') is None:
            encoding = 'UTF-8'
        if not os.path.exists(os.path.dirname(filename)):
            os.makedirs(os.path.dirname(filename), exist_ok=True)
        RotatingFileHandler.__init__(self, filename, mode, maxBytes, backupCount, encoding, delay)

        # set namer
        self.namer = self.custom_gzip_namer
        self.rotator = self.custom_gzip_rotator

    @staticmethod
    def custom_gzip_namer(name):
        """Namer for rotating log handler with gz extension.

        Args:
            name: The current name of the logfile.
        """
        return name + '.gz'

    @staticmethod
    def custom_gzip_rotator(source: str, dest: str) -> None:
        """Rotate and compress log file.

        Args:
            source: The source filename.
            dest: The destination filename.
        """
        with open(source, 'rb') as f_in:
            with gzip.open(dest, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)
        os.remove(source)


# class RotatingFileHandlerFormatter(logging.Formatter):
#     """Custom logging formatter that allows a different format depending on the logging level."""
#
#     def __init__(self):
#         """Initialize formatter parent."""
#         super().__init__(fmt='%(levelno)d: %(msg)s', datefmt=None, style='%')
#
#     def format(self, record):
#         """Format file handle log event according to logging level.
#
#         Args:
#             record (obj): The record to be logged.
#         """
#         # Replace the original format with one customized by logging level
#         self._style._fmt = self.standard_format
#         if record.levelno < 10:  # <= logging.DEBUG
#             self._style._fmt = self.trace_format
#
#         # Call the original formatter class to do the grunt work
#         result = logging.Formatter.format(self, record)
#
#         return result
#
#     @property
#     def standard_format(self):
#         """Return the standard log format"""
#         return (
#             '%(asctime)s - %(name)s - %(levelname)s - %(message)s '
#             '(%(filename)s:%(funcName)s:%(lineno)d:%(threadName)s)'
#         )
#
#     @property
#     def trace_format(self):
#         """Return the standard log format"""
#         return (
#             '%(asctime)s - %(name)s - %(levelname)s - [%(funcName)s:%(lineno)d] %(message)s '
#             '(%(filename)s:%(threadName)s)'
#         )



"""Exception that can be raised during input validation."""


class ValidationError(Exception):
    """Exception that can be raised during input validation."""

    def __init__(self, message: str):
        """Initialize a ValidationError.

        Args:
            message: A context-specific message explaining the input validation failure.
        """
        super().__init__()
        self.message = message



"""Parser for STIX AS Object.

see: https://docs.oasis-open.org/cti/stix/v2.1/csprd01/stix-v2.1-csprd01.html#_Toc16070683
"""
# standard library
from typing import Iterable, Union

# third-party
from stix2 import AutonomousSystem

# first-party
from tcex.stix.model import StixModel  # pylint: disable=cyclic-import


class StixASObject(StixModel):
    """Parser for STIX AS Object.

    see: https://docs.oasis-open.org/cti/stix/v2.1/csprd01/stix-v2.1-csprd01.html#_Toc16070683
    """

    def produce(self, tc_data: Union[list, dict], **kwargs) -> Iterable[AutonomousSystem]:
        """Produce STIX 2.0 JSON object from TC API response."""
        if isinstance(tc_data, list) and len(tc_data) > 0 and 'summary' in tc_data[0]:
            indicator_field = 'summary'
        else:
            indicator_field = 'AS Number'

        parse_map = {
            'type': 'autonomous-system',
            'spec_version': '2.1',
            'id': '@.id',
            'number': f'@."{indicator_field}"',
        }

        yield from (AutonomousSystem(**stix_data) for stix_data in self._map(tc_data, parse_map))

    def consume(self, stix_data: Union[list, dict]):
        """Produce a ThreatConnect object from a STIX 2.0 JSON object."""
        parse_map = {
            'type': 'ASN',
            'summary': '@.number',
            'xid': '@.id',
            'attributes': [{'type': 'External ID', 'value': '@.id'}],
        }

        def _convert_summary_to_str(data):
            data['summary'] = str(data.get('summary'))
            return data

        yield from map(_convert_summary_to_str, self._map(stix_data, parse_map))



"""TcEx Framework Key Value API Module"""
# standard library
from urllib.parse import quote


class KeyValueApi:
    """TcEx Key Value API Module.

    Args:
        session (request.Session): A configured requests session for TC API (tcex.session).
    """

    def __init__(self, session):
        """Initialize the Class properties."""
        self._session = session

    def create(self, key, value):
        """Create key/value pair in remote KV store.

        Args:
            key (string): The key to create in remote KV store.
            value (any): The value to store in remote KV store.

        Returns:
            (string): The response from the API call.
        """
        key = quote(key, safe='~')
        headers = {'content-type': 'application/octet-stream'}
        url = f'/internal/playbooks/keyValue/{key}'
        r = self._session.put(url, data=value, headers=headers)
        return r.content

    def read(self, key):
        """Read data from remote KV store for the provided key.

        Args:
            key (string): The key to read in remote KV store.

        Returns:
            (any): The response data from the remote KV store.
        """
        key = quote(key, safe='~')
        url = f'/internal/playbooks/keyValue/{key}'
        r = self._session.get(url)
        data = r.content
        if data is not None and not isinstance(data, str):
            data = str(r.content, 'utf-8')
        return data



"""TcEx Framework API Service module."""
# standard library
import json
import sys
import threading
import traceback
from io import BytesIO
from typing import Any

from .common_service import CommonService


class ApiService(CommonService):
    """TcEx Framework API Service module."""

    def __init__(self, tcex: object):
        """Initialize the Class properties.

        Args:
            tcex: Instance of TcEx.
        """
        super().__init__(tcex)

        # properties
        self._metrics = {'Errors': 0, 'Requests': 0, 'Responses': 0}

        # config callbacks
        self.api_event_callback = None

    @property
    def command_map(self) -> dict:
        """Return the command map for the current Service type."""
        command_map = super().command_map
        command_map.update({'runservice': self.process_run_service_command})
        return command_map

    def format_query_string(self, params: dict) -> str:
        """Convert name/value array to a query string.

        Args:
            params: The query params for the request.

        Returns:
            str: The query params reformatted as a string.
        """
        query_string = []
        try:
            for q in params:
                query_string.append(f'''{q.get('name')}={q.get('value')}''')
        except AttributeError as e:
            self.log.error(
                f'feature=api-service, event=bad-params-provided, params={params}, error="""{e})"""'
            )
            self.log.trace(traceback.format_exc())
        return '&'.join(query_string)

    def format_request_headers(self, headers: dict) -> dict:
        """Convert name/value array to a headers dict.

        Args:
            headers: The dict of key/value header data.

        Returns:
            dict: The restructured header data.
        """
        headers_ = {}
        try:
            for h in headers:
                # TODO: either support tuple or csv list of values
                # headers_.setdefault(h.get('name').lower(), []).append(h.get('value'))
                headers_.setdefault(h.get('name').lower(), str(h.get('value')))

        except AttributeError as e:
            self.log.error(
                f'feature=api-service, event=bad-headers-provided, '
                f'headers={headers}, error="""{e})"""'
            )
            self.log.trace(traceback.format_exc())
        return headers_

    def format_response_headers(self, headers: dict) -> dict:
        """Convert name/value array to a query string.

        Args:
            headers: The dict header data to be converted to key/value pairs.

        Returns:
            dict: The restructured header data.
        """
        headers_ = []
        try:
            for h in headers:
                headers_.append({'name': h[0], 'value': h[1]})
        except AttributeError as e:
            self.log.error(
                f'feature=api-service, event=bad-headers-provided, '
                f'headers={headers}, error="""{e})"""'
            )
            self.log.trace(traceback.format_exc())
        return headers_

    def process_run_service_response(self, *args, **kwargs) -> None:
        """Handle service event responses.

        ('200 OK', [('content-type', 'application/json'), ('content-length', '103')])
        """
        self.log.info('feature=api-service, event=response-received, status=waiting-for-body')
        kwargs.get('event').wait(30)  # wait for thread event - (set on body write)
        self.log.trace(f'feature=api-service, event=response, args={args}')
        try:
            status_code, status = args[0].split(' ', 1)
            response = {
                'bodyVariable': 'response.body',
                'command': 'Acknowledged',
                'headers': self.format_response_headers(args[1]),
                'requestKey': kwargs.get('request_key'),  # pylint: disable=cell-var-from-loop
                'status': status,
                'statusCode': status_code,
                'type': 'RunService',
            }
            self.log.info('feature=api-service, event=response-sent')
            self.message_broker.publish(json.dumps(response), self.args.tc_svc_client_topic)
            self.increment_metric('Responses')
        except Exception as e:
            self.log.error(
                f'feature=api-service, event=failed-creating-response-body, error="""{e}"""'
            )
            self.log.trace(traceback.format_exc())
            self.increment_metric('Errors')

    def process_run_service_command(self, message: dict) -> None:
        """Process the RunService command.

        .. code-block:: python
            :linenos:
            :lineno-start: 1

            {
              "command": "RunService",
              "apiToken": "abc123",
              "bodyVariable": "request.body",
              "headers": [ { key/value pairs } ],
              "method": "GET",
              "queryParams": [ { key/value pairs } ],
              "requestKey": "123abc",
              "userConfig": [{
                "name": "tlpExportSetting",
                "value": "TLP:RED"
              }],
            }

        Args:
            message: The message payload from the server topic.
        """
        # register config apiToken (before any logging)
        self.token.register_token(
            self.thread_name, message.get('apiToken'), message.get('expireSeconds')
        )
        self.log.info(f'feature=api-service, event=runservice-command, message="{message}"')

        # thread event used to block response until body is written
        event = threading.Event()

        # process message
        request_key: str = message.get('requestKey')
        body = None
        try:
            # read body from redis
            body_variable: str = message.pop('bodyVariable')
            if body_variable is not None:
                body: Any = self.redis_client.hget(request_key, body_variable)
                if body is not None:
                    # for API service the data in Redis is not b64 encoded
                    body = BytesIO(body)
        except Exception as e:
            self.log.error(f'feature=api-service, event=failed-reading-body, error="""{e}"""')
            self.log.trace(traceback.format_exc())
        headers: dict = self.format_request_headers(message.pop('headers'))
        method: str = message.pop('method')
        params: dict = message.pop('queryParams')
        path: str = message.pop('path')

        try:
            environ = {
                'wsgi.errors': sys.stderr,
                'wsgi.input': body,
                'wsgi.multithread': True,
                'wsgi.multiprocess': False,
                'wsgi.run_once': True,
                'wsgi.url_scheme': 'https',
                'wsgi.version': (1, 0),
                'PATH_INFO': path,
                'QUERY_STRING': self.format_query_string(params),
                'REMOTE_ADDR': message.get('remoteAddress', ''),
                # 'REMOTE_HOST': message.get('remoteAddress', ''),
                'REQUEST_METHOD': method.upper(),
                'SCRIPT_NAME': '/',
                'SERVER_NAME': '',
                'SERVER_PORT': '',
                'SERVER_PROTOCOL': 'HTTP/1.1',
            }

            # Add user config for TAXII or other service that supports the data type
            environ['user_config'] = message.get('userConfig', [])

            # add headers
            if headers.get('content-type') is not None:
                environ['CONTENT_TYPE'] = headers.pop('content-type')

            # add content length
            if headers.get('content-length') is not None:
                environ['CONTENT_LENGTH'] = headers.pop('content-length')

            for header, value in headers.items():
                environ[f'HTTP_{header}'.upper()] = value

            # make values from message available in env in camel
            # case (e.g., falcon -> req.env.get('request_url))
            for key, value in message.items():
                if key not in environ and self.tcex.utils.camel_to_snake(key) not in environ:
                    environ[self.tcex.utils.camel_to_snake(key)] = value

            self.log.trace(f'feature=api-service, environ={environ}')
            self.increment_metric('Requests')
        except Exception as e:
            self.log.error(f'feature=api-service, event=failed-building-environ, error="""{e}"""')
            self.log.trace(traceback.format_exc())
            self.increment_metric('Errors')
            return  # stop processing

        def response_handler(*args, **kwargs):  # pylint: disable=unused-argument
            """Handle WSGI Response"""
            kwargs['event'] = event  # add event to kwargs for blocking
            kwargs['request_key'] = request_key
            self.service_thread(
                name='response-handler',
                target=self.process_run_service_response,
                args=args,
                kwargs=kwargs,
            )

        if callable(self.api_event_callback):
            try:
                body_data: Any = self.api_event_callback(  # pylint: disable=not-callable
                    environ, response_handler
                )

                # process body
                body = ''
                if hasattr(body_data, 'read'):
                    body = body_data.read()
                elif isinstance(body_data, list):
                    for bd in body_data:
                        if hasattr(bd, 'read'):
                            body += bd.read()
                        elif isinstance(bd, bytes):
                            body += bd.decode()
                        elif isinstance(bd, list):
                            for b in bd:
                                self.log.error(f'unhandled type - {type(b)}')
                        else:
                            self.log.error(f'unhandled type - {type(body)}')
                            self.log.error(f'unhandled type dir - {dir(body)}')

                # write body to Redis
                self.redis_client.hset(request_key, 'response.body', body)

                # set thread event to True to trigger response
                self.log.info('feature=api-service, event=response-body-written')
                event.set()
            except Exception as e:
                self.log.error(
                    f'feature=api-service, event=api-event-callback-failed, error="""{e}""".'
                )
                self.log.trace(traceback.format_exc())
                self.increment_metric('Errors')

        # unregister config apiToken
        self.token.unregister_token(self.thread_name)



"""App Decorators Module."""
# third-party
import wrapt


class OnSuccess:
    """Set exit message on successful execution.

    This decorator will set the supplied msg as the App "exit_message". Typically and App would
    only have 1 exit message so this decorator would typically be used in App that used "actions".

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        @OnSuccess(exit_msg='Successfully processed JSON data.')
        def my_method(json_data):
            json.dumps(json_data)

    Args:
        exit_msg (str): The message to send to exit method.
    """

    def __init__(self, exit_msg=None):
        """Initialize Class properties"""
        self.exit_msg = exit_msg or 'App finished successfully.'

    @wrapt.decorator
    def __call__(self, wrapped, instance, args, kwargs):
        """Implement __call__ function for decorator.

        Args:
            wrapped (callable): The wrapped function which in turns
                needs to be called by your wrapper function.
            instance (App): The object to which the wrapped
                function was bound when it was called.
            args (list): The list of positional arguments supplied
                when the decorated function was called.
            kwargs (dict): The dictionary of keyword arguments
                supplied when the decorated function was called.

        Returns:
            function: The custom decorator function.
        """

        def completion(app, *args, **kwargs):
            """Call the function and handle any exception.

            Args:
                app (class): The instance of the App class "self".
            """
            app.exit_message = self.exit_msg
            return wrapped(*args, **kwargs)

        return completion(instance, *args, **kwargs)



"""ThreatConnect Workflow Template"""
from .api_endpoints import ApiEndpoints
from .common_case_management import CommonCaseManagement
from .common_case_management_collection import CommonCaseManagementCollection
from .filter import Filter
from .tql import TQL


class WorkflowTemplates(CommonCaseManagementCollection):
    """Workflow Template Class for Case Management Collection

    .. code-block:: python
        :linenos:
        :lineno-start: 1

        # Example of params input
        {
            'result_limit': 100,  # How many results are retrieved.
            'result_start': 10,  # Starting point on retrieved results.
            'fields': ['caseId', 'summary']  # Additional fields returned on the results
        }

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        initial_response (dict, optional): Initial data in
            Case Object for Workflow Template. Defaults to None.
        tql_filters (list, optional): List of TQL filters. Defaults to None.
        params(dict, optional): Dict of the params to be sent while
            retrieving the Workflow Event objects.
    """

    def __init__(self, tcex, initial_response=None, tql_filters=None, params=None):
        """Initialize Class properties."""
        super().__init__(
            tcex,
            ApiEndpoints.WORKFLOW_TEMPLATES,
            initial_response=initial_response,
            tql_filters=tql_filters,
            params=params,
        )

    def __iter__(self):
        """Iterate on Workflow Templates"""
        return self.iterate(initial_response=self.initial_response)

    def entity_map(self, entity):
        """Map a dict to a Workflow Template.

        Args:
            entity (dict): The Workflow Template data.

        Returns:
            CaseManagement.WorkflowTemplate: An Workflow Template Object
        """
        return WorkflowTemplate(self.tcex, **entity)

    @property
    def filter(self):
        """Return instance of FilterWorkflowTemplate Object."""
        return FilterWorkflowTemplates(ApiEndpoints.WORKFLOW_TEMPLATES, self.tcex, self.tql)


class WorkflowTemplate(CommonCaseManagement):
    """WorkflowTemplate object for Case Management.

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        active (bool, kwargs): [Read-Only] The **Active** flag for the Workflow Template.
        assignee (Assignee, kwargs): [Read-Only] The **Assignee** for the Workflow Template.
        cases (Case, kwargs): [Read-Only] The **Cases** for the Workflow Template.
        config_artifact (str, kwargs): [Read-Only] The **Config Artifact** for the Workflow
            Template.
        config_playbook (str, kwargs): [Read-Only] The **Config Playbook** for the Workflow
            Template.
        config_task (dict, kwargs): [Read-Only] The **Config Task** for the Workflow Template.
        description (str, kwargs): The **Description** for the Workflow Template.
        name (str, kwargs): [Required] The **Name** for the Workflow Template.
        target_type (str, kwargs): [Read-Only] The **Target Type** for the Workflow Template.
        version (int, kwargs): The **Version** for the Workflow Template.
    """

    def __init__(self, tcex, **kwargs):
        """Initialize Class properties."""
        super().__init__(tcex, ApiEndpoints.WORKFLOW_TEMPLATES, kwargs)
        self._active = kwargs.get('active', None)
        self._assignee = kwargs.get('assignee', None)
        self._cases = kwargs.get('cases', None)
        self._config_artifact = kwargs.get('config_artifact', None)
        self._config_playbook = kwargs.get('config_playbook', None)
        self._config_task = kwargs.get('config_task', None)
        self._description = kwargs.get('description', None)
        self._name = kwargs.get('name', None)
        self._version = kwargs.get('version', None)

    @property
    def active(self):
        """Return the parent **Active** flag for the Workflow Template."""
        return self._active

    @property
    def assignee(self):
        """Return the parent **Assignee** flag for the Workflow Template."""
        return self._assignee

    @property
    def as_entity(self):
        """Return the entity representation of the Workflow Event."""
        return {'type': 'Workflow Template', 'value': self.name, 'id': self.id}

    @property
    def cases(self):
        """Return the parent **Cases** for the Workflow Template."""
        return self.tcex.cm.cases(initial_response=self._cases)

    @property
    def config_artifact(self):
        """Return the parent **Config Artifact** for the Workflow Template."""
        return self._config_artifact

    @property
    def config_playbook(self):
        """Return the parent **Config Playbook** for the Workflow Template."""
        return self._config_playbook

    @property
    def config_task(self):
        """Return the parent **Config Task** for the Workflow Template."""
        return self._config_task

    @property
    def description(self):
        """Return the parent **Description** for the Workflow Template."""
        return self._description

    @description.setter
    def description(self, description):
        """Set the parent **Description** for the Workflow Template."""
        self._description = description

    def entity_mapper(self, entity):
        """Update current object with provided object properties.

        Args:
            entity (dict): An entity dict used to update the Object.
        """
        new_case = WorkflowTemplate(self.tcex, **entity)
        self.__dict__.update(new_case.__dict__)

    @property
    def name(self):
        """Return the parent **Name** for the Workflow Template."""
        return self._name

    @name.setter
    def name(self, name):
        """Set the parent **Name** for the Workflow Template."""
        self._name = name

    @property
    def version(self):
        """Return the parent **Version** for the Workflow Template."""
        return self._version

    @version.setter
    def version(self, version):
        """Set the parent **Version** for the Workflow Template."""
        self._version = version


class FilterWorkflowTemplates(Filter):
    """Filter Object for WorkflowTemplates"""

    def active(self, operator, active):
        """Filter Workflow Templates based on **active** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            active (bool): The active status of this template.
        """
        self._tql.add_filter('active', operator, active, TQL.Type.BOOLEAN)

    def description(self, operator, description):
        """Filter Workflow Templates based on **description** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            description (str): The description of this template.
        """
        self._tql.add_filter('description', operator, description, TQL.Type.STRING)

    def id(self, operator, id):  # pylint: disable=redefined-builtin
        """Filter Workflow Templates based on **id** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            id (int): The ID of the template.
        """
        self._tql.add_filter('id', operator, id, TQL.Type.INTEGER)

    def name(self, operator, name):
        """Filter Workflow Templates based on **name** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            name (str): The name of this template.
        """
        self._tql.add_filter('name', operator, name, TQL.Type.STRING)

    def target_id(self, operator, target_id):
        """Filter Workflow Templates based on **targetId** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            target_id (int): The ID of the target of this template.
        """
        self._tql.add_filter('targetId', operator, target_id, TQL.Type.INTEGER)

    def target_type(self, operator, target_type):
        """Filter Workflow Templates based on **targetType** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            target_type (str): The target type of this template.
        """
        self._tql.add_filter('targetType', operator, target_type, TQL.Type.STRING)

    def version(self, operator, version):
        """Filter Workflow Templates based on **version** keyword.

        Args:
            operator (enum): The operator enum for the filter.
            version (int): The version of this template.
        """
        self._tql.add_filter('version', operator, version, TQL.Type.INTEGER)



"""TcEx Framework module init file."""
# flake8: noqa
from .__metadata__ import (
    __author__,
    __author_email__,
    __description__,
    __license__,
    __package_name__,
    __url__,
    __version__,
)

try:
    from .decorators import (
        Benchmark,
        Debug,
        FailOnOutput,
        IterateOnArg,
        OnException,
        OnSuccess,
        Output,
        ReadArg,
        WriteOutput,
    )
    from .tcex import TcEx
    from .threat_intelligence import ThreatIntelligence
except ImportError as e:
    print(f'Error: {e}')
    print('Try running tclib')



"""ThreatConnect TI URL"""
# standard library
from urllib.parse import quote_plus

from ..indicator import Indicator


class URL(Indicator):
    """Unique API calls for URL API Endpoints"""

    def __init__(self, tcex, **kwargs):
        """Initialize Class Properties.

        Args:
            text (str, kwargs): [Required for Create] The URL value for this Indicator.
            active (bool, kwargs): If False the indicator is marked "inactive" in TC.
            confidence (str, kwargs): The threat confidence for this Indicator.
            date_added (str, kwargs): [Read-Only] The date timestamp the Indicator was created.
            last_modified (str, kwargs): [Read-Only] The date timestamp the Indicator was last
                modified.
            private_flag (bool, kwargs): If True the indicator is marked as private in TC.
            rating (str, kwargs): The threat rating for this Indicator.
            xid (str, kwargs): The external id for this Indicator.
        """
        super().__init__(tcex, sub_type='URL', api_entity='url', api_branch='urls', **kwargs)
        self.unique_id = kwargs.get('unique_id', kwargs.get('text'))
        self.data['text'] = self.unique_id
        if self.unique_id:
            self.unique_id = quote_plus(self.fully_decode_uri(self.unique_id))

    def can_create(self):
        """Return True if address can be created.

        If the text has been provided returns that the URL can be created, otherwise
        returns that the URL cannot be created.
        """
        return not self.data.get('text') is None

    def _set_unique_id(self, json_response):
        """Set the unique_id provided a json response.

        Args:
            json_response:
        """
        self.unique_id = quote_plus(self.fully_decode_uri(json_response.get('text', '')))



"""TcEx Runtime App Test Case"""
# standard library
import os
import subprocess
import sys

from .test_case import TestCase


class TestCaseJob(TestCase):
    """App TestCase Class"""

    run_method = 'inline'  # run service inline or a as subprocess

    @staticmethod
    def create_shelf_dir(shelf_path):
        """Create a directory in log with the context name containing the batch data."""
        if not os.path.isdir(shelf_path):
            os.makedirs(os.path.join(shelf_path, 'DEBUG'), exist_ok=True)

    def run(self):
        """Run the Playbook App.

        Returns:
            int: The exit code fo the App.
        """
        # first-party
        from run import run  # pylint: disable=no-name-in-module

        # run the app
        exit_code = 0
        try:
            run()
        except SystemExit as e:
            exit_code = e.code

        self.log.data('run', 'Exit Code', exit_code)
        return exit_code

    def run_profile(self):
        """Run an App using the profile name."""
        self.create_shelf_dir(self.profile.tc_temp_path)

        # create encrypted config file
        self.create_config(self.profile.args)

        # run the service App in 1 of 3 ways
        if self.run_method == 'inline':
            # backup sys.argv
            sys_argv_orig = sys.argv

            # clear sys.argv
            sys.argv = sys.argv[:1]

            # run the App
            exit_code = self.run()

            # restore sys.argv
            sys.argv = sys_argv_orig
        elif self.run_method == 'subprocess':
            # run the Service App as a subprocess
            app_process = subprocess.Popen(['python', 'run.py'])
            exit_code = app_process.wait()

        # run the App
        return exit_code



#!/usr/bin/env python
"""TcEx Library Builder."""
# standard library
import os
import platform
import shutil
import subprocess  # nosec
import sys
from distutils.version import StrictVersion  # pylint: disable=no-name-in-module
from urllib.parse import quote

# third-party
import colorama as c

from .bin import Bin


class Lib(Bin):
    """Install Required Modules for App.

    Args:
        _args (namespace): The argparser args Namespace.
    """

    def __init__(self, _args):
        """Initialize Class properties.

        Args:
            _args (namespace): The argparser args Namespace.
        """
        super().__init__(_args)

        # properties
        self.latest_version = None
        self.lib_directory = (
            f'lib_{sys.version_info.major}.{sys.version_info.minor}.{sys.version_info.micro}'
        )
        self.requirements_file = 'requirements.txt'
        self.static_lib_dir = 'lib_latest'
        self.use_temp_requirements_file = False

        # update tcex.json
        self.tj.update()

    def _build_command(self, python_executable, lib_dir_fq, proxy_enabled):
        """Build the pip command for installing dependencies.

        Args:
            python_executable (str): The fully qualified path of the Python executable.
            lib_dir_fq (str): The fully qualified path of the lib directory.

        Returns:
            list: The Python pip command with all required args.
        """
        exe_command = [
            os.path.expanduser(python_executable),
            '-m',
            'pip',
            'install',
            '-r',
            self.requirements_file,
            '--ignore-installed',
            '--quiet',
            '--target',
            lib_dir_fq,
        ]
        if self.args.no_cache_dir:
            exe_command.append('--no-cache-dir')

        if proxy_enabled:
            # trust the pypi hosts to avoid ssl errors
            trusted_hosts = ['pypi.org', 'pypi.python.org', 'files.pythonhosted.org']

            for host in trusted_hosts:
                exe_command.append('--trusted-host')
                exe_command.append(host)

        return exe_command

    def _configure_proxy(self):
        """Configure proxy settings using environment variables."""
        if os.getenv('HTTP_PROXY') or os.getenv('HTTPS_PROXY'):
            # TODO: is this appropriate?
            # don't change proxy settings if the OS already has them configured.
            return True

        proxy_enabled = False
        if self.args.proxy_host is not None and self.args.proxy_port is not None:
            if self.args.proxy_user is not None and self.args.proxy_pass is not None:
                proxy_user = quote(self.args.proxy_user, safe='~')
                proxy_pass = quote(self.args.proxy_pass, safe='~')

                # proxy url with auth
                proxy_url = (
                    f'{proxy_user}:{proxy_pass}@{self.args.proxy_host}:{self.args.proxy_port}'
                )
            else:
                # proxy url without auth
                proxy_url = f'{self.args.proxy_host}:{self.args.proxy_port}'

            os.putenv('HTTP_PROXY', f'http://{proxy_url}')
            os.putenv('HTTPS_PROXY', f'https://{proxy_url}')

            print(
                f'Using Proxy Server: {c.Fore.CYAN}{self.args.proxy_host}:{self.args.proxy_port}.'
            )
            proxy_enabled = True
        return proxy_enabled

    def _create_lib_latest(self):
        """Create the lib_latest symlink for App Builder."""
        if platform.system() == 'Windows':
            shutil.copytree(f'lib_{self.latest_version}', self.static_lib_dir)
        else:
            if os.path.islink(self.static_lib_dir):
                os.unlink(self.static_lib_dir)
            elif os.path.isfile(self.static_lib_dir):
                os.rmdir(self.static_lib_dir)
            os.symlink(f'lib_{self.latest_version}', self.static_lib_dir)

    def _create_temp_requirements(self):
        """Create a temporary requirements.txt.

        This allows testing again a git branch instead of pulling from pypi.
        """
        self.use_temp_requirements_file = True
        # Replace tcex version with develop branch of tcex
        with open(self.requirements_file) as fh:
            current_requirements = fh.read().strip().split('\n')

        self.requirements_file = f'temp-{self.requirements_file}'
        with open(self.requirements_file, 'w') as fh:
            new_requirements = ''
            for line in current_requirements:
                if not line:
                    continue
                if line.startswith('tcex'):
                    line = (
                        'git+https://github.com/ThreatConnect-Inc/tcex.git@'
                        f'{self.args.branch}#egg=tcex'
                    )
                # print('line', line)
                new_requirements += f'{line}\n'
            fh.write(new_requirements)

    def install_libs(self):
        """Install Required Libraries using pip."""
        # check for requirements.txt
        if not os.path.isfile(self.requirements_file):
            self.handle_error('A requirements.txt file is required to install modules.')

        # if branch arg is provide use git branch instead of pypi
        if self.args.branch is not None:
            self._create_temp_requirements()

        # default or current python version
        lib_data = [{'python_executable': sys.executable, 'lib_dir': self.lib_directory}]
        if self.tj.lib_versions:
            # overwrite default with config data
            lib_data = self.tj.lib_versions
            print(f'{c.Style.BRIGHT}Using "lib" directories defined in tcex.json file.')

        # configure proxy settings
        proxy_enabled = self._configure_proxy()

        # install all requested lib directories
        for data in lib_data:
            lib_dir = data.get('lib_dir')
            lib_dir_fq = os.path.join(self.app_path, lib_dir)

            if os.access(lib_dir_fq, os.W_OK):
                # remove lib directory from previous runs
                shutil.rmtree(lib_dir_fq)

            # replace env vars with env val in the python executable
            python_executable = os.path.expanduser(data.get('python_executable'))

            if not os.path.isfile(python_executable) and not os.path.islink(python_executable):
                print(
                    f'{c.Style.BRIGHT}{c.Fore.RED}The link Python executable ({python_executable}) '
                    f'could not be found. Skipping building lib directory for this Python version.'
                )
                continue

            print(f'Building Lib Dir: {c.Style.BRIGHT}{c.Fore.CYAN}{lib_dir_fq}')
            exe_command = self._build_command(python_executable, lib_dir_fq, proxy_enabled)

            print(f"Running: {c.Style.BRIGHT}{c.Fore.GREEN}{' '.join(exe_command)}")
            p = subprocess.Popen(
                exe_command,
                shell=False,  # nosec
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
            out, err = p.communicate()  # pylint: disable=unused-variable

            if p.returncode != 0:
                print(f'{c.Style.BRIGHT}{c.Fore.RED}FAIL')
                print(f"{c.Style.BRIGHT}{c.Fore.RED}{err.decode('utf-8')}")
                sys.exit(f"ERROR: {err.decode('utf-8')}")

            # version comparison
            try:
                python_version = lib_dir.split('_', 1)[1]
            except IndexError:
                python_version = None
                self.handle_error('Could not determine version from lib string.')

            # track the latest Python version
            if self.latest_version is None:
                self.latest_version = python_version
            elif StrictVersion(python_version) > StrictVersion(self.latest_version):
                self.latest_version = python_version

        # cleanup temp file if required
        if self.use_temp_requirements_file:
            os.remove(self.requirements_file)

        # create lib_latest
        self._create_lib_latest()



"""TcEx Framework Redis Module"""
# third-party
import redis


class RedisClient:
    """A shared REDIS client connection using a Connection Pool.

    Initialize a single shared redis.connection.ConnectionPool.
    For a full list of kwargs see https://redis-py.readthedocs.io/en/latest/#redis.Connection.

    Args:
        host (str, optional): The REDIS host. Defaults to localhost.
        port (int, optional): The REDIS port. Defaults to 6379.
        db (int, optional): The REDIS db. Defaults to 0.
        blocking_pool (bool): Use BlockingConnectionPool instead of ConnectionPool.
        errors (str, kwargs): The REDIS errors policy (e.g. strict).
        max_connections (int, kwargs): The maximum number of connections to REDIS.
        password (str, kwargs): The REDIS password.
        socket_timeout (int, kwargs): The REDIS socket timeout.
        timeout (int, kwargs): The REDIS Blocking Connection Pool timeout value.
    """

    def __init__(self, host='localhost', port=6379, db=0, blocking_pool=False, **kwargs):
        """Initialize class properties"""
        self._client = None
        pool = redis.ConnectionPool
        if blocking_pool:
            pool = redis.BlockingConnectionPool
        self.pool = pool(host=host, port=port, db=db, **kwargs)

    @property
    def client(self):
        """Return an instance of redis.client.Redis."""
        if self._client is None:
            self._client = redis.Redis(connection_pool=self.pool)
        return self._client



"""ThreatConnect TI Intrusion Set"""
from ..group import Group


class IntrusionSet(Group):
    """Unique API calls for IntrustionSet API Endpoints

    Args:
        tcex (TcEx): An instantiated instance of TcEx object.
        name (str, kwargs): [Required for Create] The name for this Group.
        owner (str, kwargs): The name for this Group. Default to default Org when not provided
    """

    def __init__(self, tcex, **kwargs):
        """Initialize Class Properties."""
        super().__init__(
            tcex,
            sub_type='Intrusion Set',
            api_entity='intrusionSet',
            api_branch='intrusionSets',
            **kwargs
        )



"""ThreatConnect STIX module"""
# standard library
from typing import Union

# third-party
from stix2 import URL

# first-party
from tcex.stix.model import StixModel


class StixURLObject(StixModel):
    """STIX Threat Actor object."""

    def consume(self, stix_data: Union[list, dict]):
        """Parse a STIX URL Object into a ThreatConnect URL Indicator."""
        mapper = {
            'type': 'URL',
            'text': '@.value',
            'xid': '@.id',
            'attributes': [{'type': 'External ID', 'value': '@.id'}],
        }

        if isinstance(stix_data, dict):
            stix_data = [stix_data]

        yield from self._map(stix_data, mapper)

    def produce(self, tc_data: Union[list, dict], **kwargs):
        """Produce STIX 2.0 JSON object from TC API response.

        .. code:: json

           {
              "type": "url",
              "spec_version": "2.1",
              "id": "url--c1477287-23ac-5971-a010-5c287877fa60",
              "value": "https://example.com/research/index.html"
            }
        """
        if isinstance(tc_data, list) and len(tc_data) > 0 and 'summary' in tc_data[0]:
            indicator_field = 'summary'
        else:
            indicator_field = 'text'

        mapper = {
            'id': '@.id',
            'value': f'@.{indicator_field}',
            'spec_version': '2.1',
            'type': 'url',
        }

        for stix_data in self._map(tc_data, mapper):
            yield URL(**stix_data)



"""Trace Logger Class"""
# standard library
import logging
from inspect import getframeinfo, stack

# Create trace logging level
logging.TRACE = logging.DEBUG - 5
logging.addLevelName(logging.TRACE, 'TRACE')


class TraceLogger(logging.Logger):
    """Add trace level to logging"""

    # supports updated call for Python 3.8
    def findCaller(
        self, stack_info=False, stack_level=1
    ):  # pylint: disable=arguments-differ,unused-argument
        """Find the caller for the current log event.

        Args:
            stack_info (bool, optional): Defaults to False.

        Returns:
            tuple: The caller stack information.
        """
        caller = None
        depth = 3
        while True:
            # search for the correct calling method
            caller = getframeinfo(stack()[depth][0])
            if caller.function != 'trace' or depth >= 6:
                break
            depth += 1

        return (caller.filename, caller.lineno, caller.function, None)

    def trace(self, msg, *args, **kwargs):
        """Set trace logging level

        Args:
            msg (str): The message to be logged.
        """
        self.log(logging.TRACE, msg, *args, **kwargs)



"""ThreatConnect Requests Session"""
# standard library
import base64
import hashlib
import hmac
import logging
import time

# third-party
import urllib3
from requests import Session, adapters, auth
from urllib3.util.retry import Retry

from ..utils import Utils

# disable ssl warning message
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


class HmacAuth(auth.AuthBase):
    """ThreatConnect HMAC Authorization"""

    def __init__(self, access_id, secret_key):
        """Initialize the Class properties."""
        super().__init__()
        self._access_id = access_id
        self._secret_key = secret_key

    def __call__(self, r):
        """Override of parent __call__ method."""
        timestamp = int(time.time())
        signature = f'{r.path_url}:{r.method}:{timestamp}'
        hmac_signature = hmac.new(
            self._secret_key.encode(), signature.encode(), digestmod=hashlib.sha256
        ).digest()
        authorization = f'TC {self._access_id}:{base64.b64encode(hmac_signature).decode()}'
        r.headers['Authorization'] = authorization
        r.headers['Timestamp'] = timestamp
        return r


class TokenAuth(auth.AuthBase):
    """ThreatConnect Token Authorization"""

    def __init__(self, token):
        """Initialize Class Properties."""
        super().__init__()
        self.token = token

    def __call__(self, r):
        """Override of parent __call__ method."""
        r.headers['Authorization'] = f'TC-Token {self.token.token}'
        return r


class TcSession(Session):
    """ThreatConnect REST API Requests Session"""

    def __init__(self, api_access_id, api_secret_key, base_url, logger=None):
        """Initialize the Class properties."""
        super().__init__()
        self.api_access_id = api_access_id
        self.api_secret_key = api_secret_key
        self.base_url = base_url.strip('/')
        self.log = logger or logging.getLogger('session')

        # properties
        self._log_curl: bool = False
        self._token = None
        self.auth = None
        self.utils = Utils()

        # Add Retry
        self.retry()

    def _configure_auth(self):
        """Return Auth property for session."""
        # Add ThreatConnect Authorization
        if self.token_available:
            # service Apps only use tokens and playbook/runtime Apps will use token if available
            self.auth = TokenAuth(self.token)
            self.log.debug('feature=tc-session, event=auth, type=token')
        elif self.api_access_id and self.api_secret_key:
            try:
                # for external Apps or testing Apps locally
                self.auth = HmacAuth(self.api_access_id, self.api_secret_key)
                self.log.debug('feature=tc-session, event=auth, type=hmac')
            except AttributeError:  # pragma: no cover
                raise RuntimeError('No valid ThreatConnect API credentials provided.')
        else:  # pragma: no cover
            raise RuntimeError('No valid ThreatConnect API credentials provided.')

    @property
    def log_curl(self) -> bool:
        """Return whether or not requests will be logged as a curl command."""
        return self._log_curl

    @log_curl.setter
    def log_curl(self, log_curl: bool):
        """Enable or disable logging curl commands."""
        self._log_curl = log_curl

    @property
    def token(self):
        """Return token."""
        return self._token

    @token.setter
    def token(self, token):
        """Set token."""
        self._token = token

    @property
    def token_available(self):
        """Return true if the current App is a service App."""
        return (
            self.token is not None
            and self.token.token is not None
            and self.token.token_expires is not None
        )

    def request(self, method, url, **kwargs):  # pylint: disable=arguments-differ
        """Override request method disabling verify on token renewal if disabled on session."""
        if self.auth is None:
            self._configure_auth()

        # accept path for API calls instead of full URL
        if not url.startswith('https'):
            url = f'{self.base_url}{url}'
        response = super().request(method, url, **kwargs)

        # don't show curl message for logging commands
        if '/v2/logs/app' not in url:
            # APP-79 - adding logging of request as curl commands
            if not response.ok or self.log_curl:
                try:
                    self.log.debug(
                        self.utils.requests_to_curl(
                            response.request, proxies=self.proxies, verify=self.verify
                        )
                    )
                except Exception:  # nosec
                    pass  # logging curl command is best effort

        self.log.debug(
            f'feature=tc-session, request-url={response.request.url}, '
            f'status_code={response.status_code}, elapsed={response.elapsed}'
        )

        return response

    def retry(self, retries=3, backoff_factor=0.3, status_forcelist=(500, 502, 504)):
        """Add retry to Requests Session

        https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#urllib3.util.retry.Retry
        """
        retries = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
        )
        # mount all https requests
        self.mount('https://', adapters.HTTPAdapter(max_retries=retries))



"""The RateLimitHandler implements request throttling based on X-RateLimit headers.

See https://tools.ietf.org/id/draft-polli-ratelimit-headers-00.html for implementation details.
"""
# standard library
import time
from typing import Optional

# third-party
from requests import PreparedRequest, Response

# first-party
from tcex.utils import Utils


class RateLimitHandler:
    """Rate-limiting implementation using X-RateLimit-<X> headers."""

    def __init__(
        self,
        limit_remaining_header: Optional[str] = 'X-RateLimit-Remaining',
        limit_reset_header: Optional[str] = 'X-RateLimit-Reset',
        remaining_threshold: Optional[int] = 0,
    ):
        """Rate-limiting implementation using X-RateLimit-<X> headers.

        See https://tools.ietf.org/id/draft-polli-ratelimit-headers-00.html for implementation
        details.  Uses X-RateLimit-Remaining and X-RateLimit-Reset to rate-limit requests by
        waiting until X-RateLimit-Reset when X-RateLimit-Remaining is less than or equal to
        remaining_threshold.

        Args:
            limit_remaining_header: Name of the header that has the limit remaining value.
            limit_reset_header: Name of the header that contains the limit reset value.
            remaining_threshold: Number of remaining requests available that will trigger a pause.
        """
        self._limit_remaining_header = limit_remaining_header
        self._limit_reset_header = limit_reset_header
        self._remaining_threshold = remaining_threshold

        self._last_limit_remaining_value = None
        self._last_limit_reset_value = None

    @property
    def last_limit_remaining_value(self) -> Optional[int]:
        """Get the last value received in the limit_remaining_header."""
        return self._last_limit_remaining_value

    @property
    def last_limit_reset_value(self) -> Optional[int]:
        """Get the last value received in the limit_reset_header."""
        return self._last_limit_reset_value

    @property
    def limit_remaining_header(self) -> str:
        """Get the name of the header that contains the remaining requests."""
        return self._limit_remaining_header

    @limit_remaining_header.setter
    def limit_remaining_header(self, limit_remaining_header):
        """Set the name of the header that contains the remaining requests."""
        self._limit_remaining_header = limit_remaining_header

    @property
    def limit_reset_header(self) -> str:
        """Get the name of the header that contains the rate limit reset timestamp."""
        return self._limit_reset_header

    @limit_reset_header.setter
    def limit_reset_header(self, limit_reset_header):
        self._limit_reset_header = limit_reset_header

    @property
    def remaining_threshold(self) -> int:
        """Get the threshold for remaining requests."""
        return self._remaining_threshold

    @remaining_threshold.setter
    def remaining_threshold(self, remaining_threshold: int):
        """Set the threshold for remaining requests."""
        self._remaining_threshold = remaining_threshold

    def post_send(self, response: Response) -> None:
        """Extract rate-limiting information from a response after a request has been sent.

        Args:
            response: The response from the request.  Should almost-never be modified.
        """
        if (
            response.ok
            and self.limit_remaining_header in response.headers
            and self.limit_reset_header in response.headers
        ):
            self._last_limit_remaining_value = int(
                response.headers.get(self.limit_remaining_header, 0)
            )
            self._last_limit_reset_value = response.headers.get(self.limit_reset_header)

    def pre_send(self, request: PreparedRequest) -> None:
        """Call before request is sent and provides an opportunity to pause for rate limiting.

        Compares rate-limit values from prior requests to determine if we should wait before sending
        request.  Calls self.sleep() if we should wait.

        Args:
            request: The request to be sent.  Should not be modified in any way.
        """
        if (
            self.last_limit_remaining_value is not None
            and self.last_limit_reset_value
            and self.last_limit_remaining_value <= self.remaining_threshold
        ):
            self.sleep(request)

    def sleep(self, request: PreparedRequest) -> None:  # pylint: disable=unused-argument
        """Sleeps to rate-limit.

        Sleeps until the time specified in X-RateLimit-Reset.

        Args:
            request:  The request that will be sent.
        """
        utils = Utils()
        wait_until = self.last_limit_reset_value
        try:
            seconds = (
                float(utils.datetime.format_datetime(wait_until, date_format='%s')) - time.time()
            )
        except RuntimeError:
            seconds = wait_until

        time.sleep(float(seconds))



"""ThreatConnect Exchange App Feature Advanced Request Module"""
# standard library
import json
from argparse import Namespace
from mimetypes import MimeTypes
from typing import Dict, List, Optional, Union

# third-party
import requests

from ..decorators import ReadArg


class AdvancedRequest:
    """App Feature Advanced Request Module

    Args:
        session (object): An instance of Requests Session object.
        tcex (object): An instance of Tcex object.
        timeout (Optional[int] = 600): The timeout value for the request.
    """

    def __init__(
        self,
        session: object,
        tcex: object,
        timeout: Optional[int] = 600,
        output_prefix: Optional[str] = None,
    ):
        """Initialize class properties."""
        self.output_prefix: str = output_prefix or tcex.ij.output_prefix
        self.session: object = session
        self.tcex: object = tcex

        # properties
        self.args: Namespace = tcex.args  # required for ReadArgs
        self.allow_redirects: bool = True
        self.data: Optional[Union[dict, str]] = None
        self.headers: dict = {}
        self.max_mb: int = 500
        self.mt: callable = MimeTypes()
        self.params: dict = {}
        self.timeout: int = timeout or 600

    @ReadArg('tc_adv_req_body')
    def configure_body(self, tc_adv_req_body: Union[bytes, str]):
        """Configure Body

        Args:
            tc_adv_req_body (Union[bytes, str]): The request body.
        """
        self.data: Union[bytes, str] = tc_adv_req_body
        if self.data is not None:
            # INT-1386
            try:
                self.data: str = self.data.encode('utf-8')
            except AttributeError:
                pass  # Binary Data

        if self.args.tc_adv_req_urlencode_body:
            try:
                self.data: dict = json.loads(self.data)
            except ValueError:  # pragma: no cover
                self.tcex.log.error('Failed loading body as JSON data.')

    @ReadArg('tc_adv_req_headers', array=True, embedded=False)
    def configure_headers(self, tc_adv_req_headers: List[Dict[str, str]]):
        """Configure Headers

        [{
            "key": "User-Agent",
            "value": "TcEx MyApp: 1.0.0",
        }]

        Args:
            tc_adv_req_headers (List[Dict[str, str]]): A dict of headers.
        """
        for header_data in tc_adv_req_headers:
            value: str = self.tcex.playbook.read(header_data.get('value'))
            self.headers[str(header_data.get('key'))] = str(value)

    @ReadArg('tc_adv_req_params', array=True, embedded=False)
    def configure_params(self, tc_adv_req_params: List[Dict[str, str]]):
        """Configure Params

        [{
            "count": "500",
            "page": "1",
        }]

        Args:
            tc_adv_req_params (List[Dict[str, str]]): A dict of Params.
        """
        for param_data in tc_adv_req_params:
            param: str = str(param_data.get('key'))
            values: str = self.tcex.playbook.read(param_data.get('value'))
            if not isinstance(values, list):
                values: list = [values]
            for value in values:
                if not value and self.args.tc_adv_req_exclude_null_params:
                    self.tcex.log.warning(
                        f'Query parameter {param} has a null/empty value '
                        'and will not be added to the request.'
                    )
                else:
                    self.params.setdefault(param, []).append(str(value))

    @ReadArg(
        'tc_adv_req_http_method',
        fail_on=[None, ''],
        fail_enabled=True,
        fail_msg='Invalid method provide for request',
        strip_values=True,
    )
    @ReadArg(
        'tc_adv_req_path',
        fail_on=[None, ''],
        fail_enabled=True,
        fail_msg='Invalid path provide for request',
        strip_values=True,
    )
    def request(self, tc_adv_req_http_method: str, tc_adv_req_path: str):
        """Make the HTTP request.

        Args:
            tc_adv_req_http_method (str): The HTTP method to use.
            tc_adv_req_path (str): The REST API endpoint/path.
        """
        # configure body
        self.configure_body(**{})

        # configure headers
        self.configure_headers(**{})

        # configure params
        self.configure_params(**{})

        # make http request
        try:
            response: object = self.session.request(
                allow_redirects=self.allow_redirects,
                data=self.data,
                headers=self.headers,
                method=tc_adv_req_http_method,
                params=self.params,
                timeout=self.timeout,
                url=tc_adv_req_path,
            )
        except requests.exceptions.RequestException as e:  # pragma: no cover
            response = None
            raise RuntimeError(f'Exception during request ({e}).')

        # write outputs as soon as they are available
        self.tcex.playbook.add_output(
            f'{self.output_prefix}.request.headers', json.dumps(dict(response.headers)), 'String'
        )
        self.tcex.playbook.add_output(
            f'{self.output_prefix}.request.ok', str(response.ok).lower(), 'String'
        )
        self.tcex.playbook.add_output(
            f'{self.output_prefix}.request.reason', response.reason, 'String'
        )
        self.tcex.playbook.add_output(
            f'{self.output_prefix}.request.status_code', response.status_code, 'String'
        )
        self.tcex.playbook.add_output(
            f'{self.output_prefix}.request.url', response.request.url, 'String'
        )

        # get response size
        response_bytes: int = len(response.content)
        response_mb: float = response_bytes / 1000000
        self.tcex.log.info(f'Response MB: {response_mb}')
        if response_mb > self.max_mb:  # pragma: no cover
            raise RuntimeError('Download was larger than maximum supported 500 MB.')

        # write content after size validation
        self.tcex.playbook.add_output(
            f'{self.output_prefix}.request.content', response.text, 'String'
        )
        self.tcex.playbook.add_output(
            f'{self.output_prefix}.request.content.binary', response.content, 'Binary'
        )

        # fail if fail_on_error is selected and not ok
        if self.args.tc_adv_req_fail_on_error and not response.ok:
            raise RuntimeError(f'Failed for status ({response.status_code})')

        return response



"""ThreatConnect STIX module"""
# standard library
from typing import Union

# third-party
from stix2 import IPv6Address

# first-party
from tcex.stix.model import StixModel  # pylint: disable=cyclic-import


class StixIPBase(StixModel):
    """Base class for IPv4 and IPv6 objects."""

    def do_consume(self, stix_data: Union[list, dict], full_block_size: int):
        """Convert a STIX object to a TC Address or CIDR.

        Args:
            stix_data: the data to convert
            full_block_size: the size of a full block in the respective CIDR
        """
        mapper_ip = {
            'type': 'Address',
            'ip': '@.value',
            'xid': '@.id',
            'attributes': [{'type': 'External ID', 'value': '@.id'}],
        }

        mapper_cider = {
            'type': 'CIDR',
            'block': '@.value',
            'xid': '@.id',
            'attributes': [{'type': 'External ID', 'value': '@.id'}],
        }

        if isinstance(stix_data, dict):
            stix_data = [stix_data]

        for data in stix_data:
            cidr_parts = data.get('value', '').split('/')
            cidr_suffix = cidr_parts[1] if len(cidr_parts) > 1 else str(full_block_size)
            if cidr_suffix == str(full_block_size):
                yield from self._map(data, mapper_ip)
            else:
                yield from self._map(data, mapper_cider)


class StixIPv4Object(StixIPBase):
    """STIX Threat Actor object."""

    def consume(self, stix_data: Union[list, dict]):
        """Convert STIX IPv4-addr Cyber Observables to ThreatConnect indicators.

        Args:
            stix_data: One or more STIX ipv4-addr cyber observables.

        Yields:
            ThreatConnect addresses or CIDRs.
        """
        yield from super().do_consume(stix_data, 32)

    def produce(self, tc_data: Union[list, dict], **kwargs):
        """Produce STIX 2.0 JSON object from TC API response.

        .. code:: json

            {
              "type": "ipv4-addr",
              "spec_version": "2.1",
              "id": "ipv4-addr--ff26c055-6336-5bc5-b98d-13d6226742dd",
              "value": "198.51.100.3"
            }
        """
        if isinstance(tc_data, dict):
            tc_data = [tc_data]

        for data in tc_data:
            if 'ip' in data or data.get('type') == 'Address':
                yield from self._produce_address(data)
            else:
                yield from self._produce_cidr(data)

    def _produce_cidr(self, tc_data: dict):
        indicator_field = 'summary' if 'summary' in tc_data else 'block'

        ip_type = 'ipv6' if ':' in tc_data.get(indicator_field, '') else 'ipv4'

        yield from self._map(
            tc_data,
            {
                'id': '@.id',
                'value': f'@.{indicator_field}',
                'spec_version': '2.1',
                'type': f'{ip_type}-addr',
            },
        )

    def _produce_address(self, tc_data: dict):
        indicator_field = 'summary' if 'summary' in tc_data else 'ip'

        ip_type = 'ipv6' if ':' in tc_data.get(indicator_field, '') else 'ipv4'

        yield from self._map(
            tc_data,
            {
                'id': '@.id',
                'value': f'@.{indicator_field}',
                'spec_version': '2.1',
                'type': f'{ip_type}-addr',
            },
        )


class StixIPv6Object(StixIPBase):
    """STIX Threat Actor object."""

    def consume(self, stix_data: Union[list, dict]):
        """Convert STIX IPv6-addr Cyber Observables to ThreatConnect indicators.

        Args:
            stix_data: One or more STIX ipv6-addr cyber observables.

        Yields:
            ThreatConnect addresses or CIDRs.
        """
        yield from super().do_consume(stix_data, 128)

    def produce(self, tc_data: Union[list, dict], **kwargs):
        """Produce STIX 2.0 JSON object from TC API response.

        .. code:: json

            {
              "type": "ipv6-addr",
              "spec_version": "2.1",
              "id": "ipv6-addr--1e61d36c-a16c-53b7-a80f-2a00161c96b1",
              "value": "2001:0db8:85a3:0000:0000:8a2e:0370:7334"
            }
        """
        if isinstance(tc_data, list) and len(tc_data) > 0 and 'summary' in tc_data[0]:
            indicator_field = 'summary'
        else:
            indicator_field = 'ip'

        mapper = {
            'id': '@.id',
            'value': f'@.{indicator_field}',
            'spec_version': '2.1',
            'type': 'ipv6-addr',
        }

        yield from (IPv6Address(**stix_data) for stix_data in self._map(tc_data, mapper))



"""TcEx Framework Service module"""
# standard library
import threading
import time
from typing import Optional

# third-party
from requests import Session, exceptions
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from ..utils import Utils


def retry_session(retries=3, backoff_factor=0.8, status_forcelist=(500, 502, 504)):
    """Add retry to Requests Session.

    https://urllib3.readthedocs.io/en/latest/reference/urllib3.util.html#urllib3.util.retry.Retry
    """
    session = Session()
    retries = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    # mount all https requests
    session.mount('https://', HTTPAdapter(max_retries=retries))
    return session


class Tokens:
    """Service methods for customer Service (e.g., Triggers)."""

    def __init__(self, token_url: str, sleep_interval: int, verify: bool, logger: object):
        """Initialize the Class properties.

        Args:
            token_url: The ThreatConnect URL for token renewal.
            sleep_interval: Token monitor sleep interval.
            verify: A boolean to enable/disable SSL verification.
            logger: An pre-configured instance of a logger.
        """
        self.token_url = token_url
        self.sleep_interval = sleep_interval
        self.log = logger
        self.verify = verify

        # properties
        self.lock = threading.Lock()
        # session with retry for token renewal
        self.session: Session = retry_session()
        # shutdown boolean
        self.shutdown = False
        # token map for storing keys -> tokens -> threads
        self.token_map = {}
        # amount of seconds to pad before token renewal
        self.token_window = 60
        self.utils = Utils

        # start token renewal process
        self.token_renewal()

    @property
    def key(self) -> str:
        """Return the current key"""
        key = 'MainThread'  # default Python parent thread name
        if self.thread_name in self.token_map:
            # for Job, Playbook, and ApiService Apps the key is the thread name.
            key: str = self.thread_name
        elif self.trigger_id in self.token_map:
            key: str = self.trigger_id
        return key

    def register_token(self, key: str, token: str, expires: int) -> None:
        """Register a token.

        Args:
            key: The key to use to identify a token. Typically a thread name or a config id.
            token: The ThreatConnect API token.
            expires: The token expiration timestamp.
        """
        if token is None or expires is None:  # pragma: no cover
            self.log.error(
                'feature=token, event=invalid-token-data, '
                f'token="{self.utils.printable_cred(token, 10)}", '
                f'expires={expires}.'
            )
            return

        self.token_map[key] = {'token': token, 'token_expires': int(expires)}
        self.log.info(
            f'feature=token, action=token-register, key={key}, '
            f'token="{self.utils.printable_cred(token, 10)}", '
            f'expiration={expires}'
        )

    def renew_token(self, token: str) -> None:
        """Renew expired ThreatConnect Token.

        This method will renew a token and update the token_map with new token and expiration.

        Args:
            token: The ThreatConnect API token.
        """
        api_token_data = {}
        self.log.in_token_renewal = True  # pause API logging

        # log token information
        try:
            params = {'expiredToken': token}
            url = f'{self.token_url}/appAuth'
            r = self.session.get(url, params=params, verify=self.verify)

            if not r.ok:
                err_reason = r.text or r.reason
                err_msg = (
                    f'feature=token, event=token-retry-error, status_code={r.status_code}, '
                    f'api-message={err_reason}, '
                    f'token="{self.utils.printable_cred(token, 10)}".'
                )
                self.log.error(err_msg)
                raise RuntimeError(1042, err_msg)
        except exceptions.SSLError:  # pragma: no cover
            raise RuntimeError('Token renewal failed with an SSL Error.')

        # process response for token
        try:
            api_token_data = r.json()
        except (AttributeError, ValueError) as e:  # pragma: no cover
            raise RuntimeError(f'Token renewal failed ({e}).')
        finally:
            self.log.in_token_renewal = False

        return api_token_data

    @property
    def thread_name(self) -> str:
        """Return the current thread name."""
        return threading.current_thread().name

    @property
    def token(self) -> Optional[str]:
        """Return token for current thread."""
        return self.token_map.get(self.key, {}).get('token')

    @token.setter
    def token(self, token: str) -> None:
        """Set token for current thread."""
        self.token_map.setdefault(self.key, {})['token'] = token

    @property
    def token_expires(self) -> Optional[int]:
        """Return token_expires for current thread."""
        return self.token_map.get(self.key, {}).get('token_expires')

    @token_expires.setter
    def token_expires(self, expires) -> None:
        """Set token expires for current thread."""
        self.token_map.setdefault(self.key, {})['token_expires'] = int(expires)

    def token_renewal(self) -> None:
        """Start token renewal monitor thread."""
        t = threading.Thread(name='token-renewal', target=self.token_renewal_monitor, daemon=True)
        t.start()

    def token_renewal_monitor(self) -> None:
        """Monitor token expiration and renew when required."""
        self.log.debug('feature=token, event=renewal-monitor-started')
        while True:
            for key, token_data in dict(self.token_map).items():
                # calculate the time left to sleep
                sleep_seconds = (
                    token_data.get('token_expires') - int(time.time()) - self.token_window
                )
                self.log.debug(
                    '''feature=token, '''
                    f'''event=token-status, key={key}, '''
                    f'''token="{self.utils.printable_cred(token_data.get('token'), 10)}", '''
                    f'''expires={token_data.get('token_expires')}, '''
                    f'''sleep-seconds={sleep_seconds}'''
                )

                if sleep_seconds > 0:
                    continue

                # renew token data
                with self.lock:
                    try:
                        api_token_data = self.renew_token(token_data.get('token'))
                        self.token_map[key]['token'] = api_token_data['apiToken']
                        self.token_map[key]['token_expires'] = int(
                            api_token_data['apiTokenExpires']
                        )
                        self.log.info(
                            f'''feature=token, action=token-renewed, key={key}, token='''
                            f'''"{self.utils.printable_cred(api_token_data['apiToken'], 10)}", '''
                            f'''expires={api_token_data['apiTokenExpires']}'''
                        )
                    except RuntimeError as e:
                        self.log.error(e)
                        try:
                            del self.token_map[key]
                            self.log.error(f'feature=token, event=token-removal-failure, key={key}')
                        except KeyError:  # pragma: no cover
                            pass
            time.sleep(self.sleep_interval)
            if self.shutdown:
                break

    @property
    def trigger_id(self) -> Optional[int]:
        """Return the current trigger_id."""
        trigger_id = None
        if hasattr(threading.current_thread(), 'trigger_id'):
            trigger_id = threading.current_thread().trigger_id
        if trigger_id is not None:
            trigger_id = int(trigger_id)
        return trigger_id

    def unregister_token(self, key: str) -> None:
        """Unregister a token.

        Args:
            key (str): The key used to identify a token.
        """
        try:
            del self.token_map[key]
            self.log.info(f'feature=token, action=token-unregister, key={key}')
        except KeyError:
            pass
